import inspect
from typing import Callable, Union


def extract_var_names(obj: Union[str, Callable]) -> dict:
    """Extract variables from class or function."""
    return {key: value.default for key, value in inspect.signature(obj).parameters.items()}



# import fnmatch
# import os
# import re
# import shutil
# import subprocess
# import sys
#
# import docutils.core
# from setuptools import Extension, find_packages
# from setuptools.command.build_ext import build_ext
#
# # Convert distutils Windows platform specifiers to CMake -A arguments
# PLAT_TO_CMAKE = {
#     "win32": "Win32",
#     "win-amd64": "x64",
#     "win-arm32": "ARM",
#     "win-arm64": "ARM64",
# }
#
#
# # A CMakeExtension needs a sourcedir instead of a file list.
# # The name must be the _single_ output extension from the CMake build.
# # If you need multiple extensions, see scikit-build.
# class CMakeExtension(Extension):
#     def __init__(self, name, sourcedir=""):
#         Extension.__init__(self, name, sources=[])
#         self.sourcedir = os.path.abspath(sourcedir)
#
#
# class CMakeBuild(build_ext):
#     def __init__(self, dist):
#         self.distribution = None
#         super().__init__(dist)
#
#     def build_extension(self, ext):  # noqa: C901
#         ext_dir = self.get_ext_fullpath(ext.name)
#         ext_dir = os.path.dirname(ext_dir)
#         ext_dir = os.path.abspath(ext_dir)
#
#         # required for auto-detection & inclusion of auxiliary "native" libs
#         if not ext_dir.endswith(os.path.sep):
#             ext_dir += os.path.sep
#
#         debug = self.debug or int(os.environ.get("DEBUG", 0))
#         cfg = "Debug" if debug else "Release"
#
#         # CMake lets you override the generator - we need to check this.
#         # Can be set with Conda-Build, for example.
#         cmake_generator = os.environ.get("CMAKE_GENERATOR", "")
#
#         # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
#         # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
#         # from Python.
#         cmake_args = [
#             f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={ext_dir}",
#             f"-DPYTHON_EXECUTABLE={sys.executable}",
#             f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
#         ]
#         build_args = []
#         # Adding CMake arguments set as environment variable
#         # (needed e.g. to build for ARM OSx on conda-forge)
#         if "CMAKE_ARGS" in os.environ:
#             cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]
#
#         # In this example, we pass in the version to C++. You might not need to.
#         cmake_args += [f"-DVERSION_INFO={self.distribution.get_version()}"]
#
#         if self.compiler.compiler_type == "msvc":
#
#             # Single config generators are handled "normally"
#             single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})
#
#             if not single_config:
#                 # CMake allows an arch-in-generator style for backward compatibility
#                 contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})
#
#                 if not contains_arch:
#                     cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]
#
#                 cmake_args += [f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={ext_dir}"]
#                 build_args += ["--config", cfg]
#
#         elif not cmake_generator:
#             try:
#                 import ninja  # noqa: F401
#
#                 cmake_args += ["-GNinja"]
#             except ImportError:
#                 pass
#
#         if sys.platform.startswith("darwin"):
#             if archs := re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", "")):
#                 cmake_args += [f'-DCMAKE_OSX_ARCHITECTURES={";".join(archs)}']
#
#         # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
#         # across all generators.
#         if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ and hasattr(self, "parallel") and self.parallel:
#             # CMake 3.12+ only.
#             build_args += [f"-j{self.parallel}"]
#
#         build_temp = os.path.join(self.build_temp, ext.name)
#         if not os.path.exists(build_temp):
#             os.makedirs(build_temp)
#
#         subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, cwd=build_temp)
#         subprocess.check_call(["cmake", "--build", "."] + build_args, cwd=build_temp)
#
#
# # Setuptools bug causes collision between pypy and cpython artifacts
# # if os.path.exists("build"):
# #     shutil.rmtree("build")
#
# docutils.core.publish_file(source_path="README.rst", destination_path="README.html", writer_name="html")
#
# with open("README.html") as fh:
#     long_description = fh.read()
#
#
# def build(setup_kwargs):
#     ext_modules = [CMakeExtension("_cornflakes", "inst/_cornflakes")]
#     setup_kwargs.update(
#         {
#             "ext_modules": ext_modules,
#             "cmdclass": {"build_ext": CMakeBuild},
#             "zip_safe": False,
#             "long_description": long_description,
#             "long_description_content_type": "text/html",
#             # "packages": find_packages("src"),
#             # "package_dir": {"": "src"},
#         }
#     )
from glob import glob
import os

import pybind11
from pybind11.setup_helpers import Pybind11Extension, build_ext

__version__ = "3.0.4"  # <<COOKIETEMPLE_FORCE_BUMP>>
with open("cornflakes/__init__.py") as f:
    while line := f.readline():
        if "__version__" in line:
            __version__ = eval(line.split("=").pop().strip())

os.environ["VERSION_INFO"] = __version__

inst_path = "inst"
external_path = f"{inst_path}/ext"
path = f"{inst_path}/cornflakes"
files = [
    f
    for f in [
        *glob(f"{path}/*/**", recursive=True),
        *glob(f"{path}/*", recursive=True),
        *glob(f"{external_path}/hash-library/*", recursive=True),
    ]
    if "test" not in f
    if os.path.splitext(f)[1] == ".cpp"
]

ext_paths = [external_path, f"{external_path}/pybind11/include", f"{external_path}/rapidjson/include/rapidjson"]


def build(setup_kwargs):
    ext_modules = [
        Pybind11Extension("_cornflakes", [*files], include_dirs=[pybind11.get_include(), path, *ext_paths], cxx_std=17),
    ]
    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmdclass": {"build_ext": build_ext},
            "zip_safe": False,
        }
    )



import datetime
from typing import Optional, cast


def _check_milliseconds_field(millisecond) -> None:
    if not 0 <= millisecond <= 999:
        raise ValueError("microsecond must be in 0..999", millisecond)


class DatetimeMS(datetime.datetime):
    """Class for datetime with milliseconds instead of microseconds."""

    def __new__(cls, year, month=None, day=None, hour=0, minute=0, second=0, millisecond=0, tzinfo=None, *, fold=0):
        """Class for datetime with milliseconds instead of microseconds.

        :param year: int
        :param month: int
        :param day: int
        :param hour: int
        :param minute: int
        :param second: int
        :param millisecond: int
        :param tzinfo: datetime.tzinfo
        :param fold: int

        :return: Datetime instance
        :rtype: DatetimeMS
        """
        _check_milliseconds_field(millisecond)
        self: datetime.datetime = datetime.datetime.__new__(
            cls, year, month, day, hour, minute, second, millisecond * 1000, tzinfo
        )

        return self

    def __repr__(self):
        """Convert to formal string, for repr()."""
        datetime_list = [
            self.year,
            self.month,
            self.day,  # These are never zero
            self.hour,
            self.minute,
            self.second,
            self.millisecond,
        ]
        if datetime_list[-1] == 0:
            del datetime_list[-1]
        if datetime_list[-1] == 0:
            del datetime_list[-1]
        datetime_str = f"datetime_ms({', '.join(map(str, datetime_list))})"
        if self.tzinfo is not None:
            if datetime_str[-1:] == ")":
                datetime_str = datetime_str[:-1] + ", tzinfo=%r" % self.tzinfo + ")"
        if self.fold:
            if datetime_str[-1:] == ")":
                datetime_str = datetime_str[:-1] + ", fold=1)"
        return datetime_str

    @property
    def millisecond(self) -> int:
        """Milliseconds of datetime.

        :return: milliseconds
        :rtype: int
        """
        return int(self.microsecond / 1000)

    def __str__(self):
        """Convert to string, for str()."""
        return self.isoformat(sep=" ", timespec="milliseconds")[:-3]


def datetime_ms(
    year,
    month: Optional[int] = None,
    day: Optional[int] = None,
    hour: Optional[int] = 0,
    minute: Optional[int] = 0,
    second: Optional[int] = 0,
    millisecond: Optional[int] = 0,
    tzinfo: Optional[datetime.tzinfo] = None,
) -> DatetimeMS:
    """Create Instance of :meth:`cornflakes.common.DatetimeMS`."""
    return cast(DatetimeMS, DatetimeMS(year, month, day, hour, minute, second, millisecond, tzinfo))  # camelcase


class UnquotedString(str):
    """String Wrapper with overwritten repr Function to not quote."""

    def __repr__(self):
        """String Wrapper with overwritten repr Function to not quote."""
        return self


def unquoted_string(x: str):
    """Create Instance of :meth:`cornflakes.common.UnquotedString`."""
    return UnquotedString(x)



"""cornflakes config generation."""
from dataclasses import is_dataclass
import importlib
import inspect
import logging
import os
import sysconfig
from types import ModuleType
from typing import Dict, List, Optional, Union

import cornflakes.builder.config_template
from cornflakes.common import unquoted_string
from cornflakes.decorator import ConfigArguments, Loader, field
from cornflakes.decorator.config import config_files, config_group, is_config


def generate_group_module(  # noqa: C901
    source_module: Union[ModuleType, str],
    source_config: Optional[Union[Dict[str, Union[List[str], str]], List[str], str]] = None,
    target_module_file: Optional[str] = None,
    class_name: Optional[str] = None,
    loader: Loader = Loader.FILE_LOADER,
    *args,
    **kwargs,
):
    """Module with function to generate automatically config group module."""
    declaration = []
    ini_config_objects = {}
    imports = []
    extra_imports = []
    files = kwargs.get(ConfigArguments.files.name, [])
    files = files if isinstance(files, list) else [files]

    if ConfigArguments.files.name not in kwargs:
        kwargs.update({ConfigArguments.files.name: source_config})

    if not target_module_file:
        target_module_file = f'{source_module.__name__.replace(".", "/")}/default.py'

    with open(cornflakes.builder.config_template.__file__) as file:
        template = file.read()

    if class_name:
        template = template.replace("class Config", f"class {class_name}")
        template = template.replace('["Config"]', f'["{class_name}"]')

    # Write Template to prevent import errors
    with open(target_module_file, "w") as file:
        file.write(template)

    if isinstance(source_module, str):
        source_module = importlib.import_module(source_module)

    for cfg_name, cfg_class in inspect.getmembers(source_module):
        if is_dataclass(cfg_class) and inspect.isclass(cfg_class) and is_config(cfg_class):
            cfg = getattr(cfg_class, str(loader.value))(files=source_config)
            ini_config_objects.update(cfg)
            imports.append(cfg_name)
            files.extend([file for file in config_files(cfg_class) if file and file not in files])

    if ConfigArguments.filter_function.name in kwargs:
        filter_function = kwargs.pop(ConfigArguments.filter_function.name)
        kwargs[ConfigArguments.filter_function.name] = unquoted_string(filter_function.__name__)
        extra_imports.append(f"from {filter_function.__module__} import {filter_function.__name__}")

    logging.debug(f"Found configs: {imports}")

    declaration.extend(
        [
            (
                f"{cfg_name}: List[{cfg[0].__class__.__name__}] = "
                f"{field.__name__}(default_factory={cfg.__class__.__name__})"
                if isinstance(cfg, list)
                else f"{cfg_name}: {cfg.__class__.__name__} = {cfg.__class__.__name__}()"
            )
            for cfg_name, cfg in ini_config_objects.items()
        ]
    )

    extra_imports.extend(
        [f"from {field.__module__} import {field.__name__}", "from typing import List"] if declaration else []
    )

    if args or kwargs:
        kwargs.update({key: repr(value).replace("'", '"') for key, value in kwargs.items()})
        template = template.replace(
            f"@{config_group.__name__}",
            (
                f"@{config_group.__name__}("
                f"{', '.join([*args, *[f'{key}={value}' for key, value in kwargs.items()]])})"
            ),
        )

    template = template.replace(
        "from",
        (
            f"""{'''
'''.join(extra_imports)}\n\nfrom {source_module.__name__} import {
            ''', '''.join(imports)}\nfrom"""
        ),
    )
    template = template.replace(
        "# modules",
        f"""{'''
    '''.join(declaration)}""",
        1,
    )

    with open(target_module_file, "w") as file:
        file.write(template)

    if "black" in os.listdir(sysconfig.get_paths()["purelib"]):
        # fix format
        # source_config = " ".join(source_config) if isinstance(source_config, list) else source_config
        os.system(f"black {target_module_file}")  # noqa: S605
    if "isort" in os.listdir(sysconfig.get_paths()["purelib"]):
        os.system(f"isort {target_module_file}")  # noqa: S605



"""cornflakes CLI.
__________________________________

.. currentmodule:: cornflakes.cli

.. autosummary::
   :toctree: _generate

"""  # noqa: RST303 D205
from cornflakes import config as config_wrapper
from cornflakes.click import command, auto_option


@config_wrapper
class Test:
    """Test CLI Config."""

    test: str = "blub"


@command("test")
@auto_option(Test)
def test(config):
    """Test CLI."""
    print(config)



from click import Argument


class RichArg(Argument):
    """A custom argument implementation of click.

    Argument class in order to provide a short help message for each argument of a command.
    """

    def __init__(self, *args, **kwargs):
        """Init Method for Rich Argument Class."""
        self.help = kwargs.pop("help", None)
        super().__init__(*args, **kwargs)



import logging

from cornflakes.click.options._global import global_option
from cornflakes.logging import setup_logging


@global_option(
    ["-v", "--verbose"],
    is_flag=True,
    help="Base logging level is set to logging.DEBUG.",
)
def verbose_option(verbose):
    """Default Option for verbose logging."""
    setup_logging(default_level=logging.DEBUG if verbose else None, force=True)



from functools import wraps
from inspect import isclass
import logging
import logging.config
import os
from types import FunctionType
from typing import Any, Callable, Optional, Protocol, Union

from rich.logging import RichHandler
import yaml


def setup_logging(  # noqa: C901
    default_path: str = "logging.yaml",
    default_level: Optional[int] = None,
    env_key: str = "LOG_CFG",
    force: bool = False,
    **kwargs,
):
    """Setup logging configuration.

    :param force: Overwrite current log-level
    :param default_path: Default path to logging config file.
    :param default_level: Default log-level (Logging.INFO).
    :param env_key: Environment key to use for logging configuration.
    :param kwargs: arguments to pass to rich_handler
    """
    if value := os.getenv(env_key, None):
        default_path = value

    if os.path.exists(default_path):
        with open(default_path) as f:
            config = yaml.safe_load(f.read())
            if default_level and force:
                for handler in config["root"]["handlers"]:
                    config["handlers"][handler]["level"] = default_level or logging.root.level
            logging.config.dictConfig(config)
    else:
        if not any([isinstance(handler, RichHandler) for handler in logging.root.handlers]):
            rich_handler_kargs: Any = {"log_time_format": "[%Y-%m-%d %H:%M:%S.%f]"}
            rich_handler_kargs.update(
                {key: value for key, value in kwargs.items() if key in RichHandler.__init__.__code__.co_varnames}
            )
            rich_handler = RichHandler(rich_tracebacks=True, **rich_handler_kargs)
            rich_handler.setFormatter(fmt=logging.Formatter("%(name)s - %(funcName)s() - %(message)s"))
            rich_handler.setLevel(default_level or logging.root.level)
            # logging.root.handlers.clear()
            logging.root.addHandler(rich_handler)
        for handler in logging.root.handlers:
            if hasattr(handler, "setLevel"):
                handler.setLevel(default_level or logging.root.level)
        for logger in logging.root.manager.loggerDict.values():
            if isinstance(logger, logging.Logger) and hasattr(logger, "__cornflakes__"):
                logger.setLevel(default_level or logging.root.level)
        if isinstance(logging.root, logging.Logger):
            logging.root.setLevel(default_level or logging.root.level)


class LoggerMetaClass(type):
    """LoggerMetaClass used for  metaclass."""

    def __new__(mcs, classname, bases, class_dict):  # noqa: N804
        """Method __new__ for LoggerMetaClass."""
        new_class_dict = {}
        for attribute_name, attribute in class_dict.items():
            if isinstance(attribute, FunctionType):
                # replace it with a wrapped version
                attribute = attach_log(attribute)
            new_class_dict[attribute_name] = attribute
        mcs.logger = logging.getLogger(classname)
        return type.__new__(mcs, classname, bases, new_class_dict)


class LoggerProtocol(Protocol):
    """LoggerProtocol used for Type Annotation."""

    logger: logging.Logger


def __wrap_class(
    w_obj,
    log_level: Optional[int] = None,
):
    w_obj.logger = logging.getLogger(f"{w_obj.__module__}.{w_obj.__name__}")
    w_obj.logger.__cornflakes__ = True
    w_obj.logger.setLevel(log_level or logging.root.level)

    if w_obj.logger.level == logging.DEBUG:
        for attribute_name, attribute in w_obj.__dict__.items():
            if isinstance(attribute, FunctionType):
                # replace it with a wrapped version
                attribute = attach_log(
                    obj=attribute,
                    log_level=log_level,
                )
                setattr(w_obj, attribute_name, attribute)
    return w_obj


def __wrap_function(
    w_obj,
    log_level: Optional[int] = None,
):
    logger = logging.getLogger(f"{w_obj.__module__}.{w_obj.__qualname__}")
    logger.__cornflakes__ = True
    logger.setLevel(log_level or logging.root.level)
    if logger.level != logging.DEBUG:
        return w_obj

    @wraps(w_obj)
    def wrapper(*args, **kwargs):
        call_signature = ", ".join([repr(a) for a in args] + [f"{k}={v!r}" for k, v in kwargs.items()])
        logger.debug(f"function {w_obj.__name__} called with args {call_signature}")
        try:
            return w_obj(*args, **kwargs)
        except Exception as e:
            logger.exception(f"Exception raised in {w_obj.__name__}. exception: {str(e)}")
            raise e

    return wrapper


def attach_log(
    obj=None,
    log_level: Optional[int] = None,
    default_level: Optional[int] = None,
    default_path: str = "logging.yaml",
    env_key: str = "LOG_CFG",
):
    """Function decorator to attach Logger to functions.

    :param obj: Logger function or class to attach the logging to.
    :param log_level: log-level for the current object logging.
    :param default_path: Default path to logging config file.
    :param default_level: Default log-level (Logging.INFO).
    :param env_key: Environment key to use for logging configuration.

    :returns: Object with attached logging instance
    """
    if default_level:
        setup_logging(default_path, default_level, env_key, force=True)

    def obj_wrapper(w_obj) -> Union[LoggerProtocol, Callable[..., Any]]:
        if isclass(w_obj):
            return __wrap_class(w_obj, log_level)

        if callable(w_obj):
            return __wrap_function(w_obj, log_level)

        return obj

    if not obj:
        return obj_wrapper
    return obj_wrapper(obj)



"""cornflakes.click.rich Module."""
from click import group as click_group
from click import command as click_command
from click import argument as click_argument

from cornflakes.common import patch_module
from cornflakes.click.rich._rich_config import RichConfig
from cornflakes.click.rich._rich_group import RichGroup
from cornflakes.click.rich._rich_argument import RichArg
from cornflakes.click.rich._rich_global_option_wrapper import rich_global_option_wrapper
from typing import Callable, Union

from cornflakes.click.rich._rich_command import RichCommand

F = Callable[..., Union[RichCommand, RichGroup, RichArg]]


def group(*args, cls=RichGroup, **kwargs) -> F:  # type: ignore
    """Group decorator function.

    Defines the group() function so that it uses the RichGroup class by default.
    """
    return rich_global_option_wrapper(click_group, *args, cls=cls, **kwargs)


def command(*args, cls=RichCommand, **kwargs) -> F:  # type: ignore
    """Command decorator function.

    Defines the command() function so that it uses the RichCommand class by default.
    """
    return rich_global_option_wrapper(click_command, *args, cls=cls, **kwargs)


def argument(*args, cls=RichArg, **kwargs) -> F:  # type: ignore
    """Command decorator function.

    Defines the command() function so that it uses the RichCommand class by default.
    """
    return click_argument(*args, cls=cls, **kwargs)


__all__ = ["RichConfig", "RichGroup", "RichCommand", "RichArg", "command", "group", "argument"]

patch_module(globals())



import logging
from typing import Optional

from cornflakes import ini_load
from cornflakes.common import type_to_str
from cornflakes.decorator._types import Config
from cornflakes.decorator.config._helper import get_not_ignored_slots, is_config
from cornflakes.decorator.config._load_config import create_file_loader
from cornflakes.decorator.config._write_config import write_config


def _parse_config_list(cfg, cfg_name: str, title: str):
    _ini_bytes = bytearray()
    has_list = isinstance(cfg, list)
    if is_config(cfg) and not has_list:
        return cfg.to_ini_bytes(cfg_name)
    elif has_list:
        for n, sub_cfg in enumerate(cfg):
            cfg_name = f"{cfg_name}_{n}"
            if is_config(sub_cfg) and hasattr(sub_cfg, "section_name"):
                cfg_name = sub_cfg.section_name
            _ini_bytes.extend(_parse_config_list(sub_cfg, cfg_name, title))
        return _ini_bytes
    else:
        logging.warning(f"The Value {cfg_name} of {title} be in a child config class!")
        return b""


def to_ini_bytes(
    self, title: Optional[str] = None
) -> Optional[bytearray]:  # TODO: implement more type_to_str feature -> date format etc.
    """Method to write an instance of the main config class of the module into a ini bytearray."""
    _ini_bytes = bytearray()
    has_lists = isinstance(self, list)
    if is_config(self) and not has_lists:
        _ini_bytes.extend(bytes(f"[{title}]\n", "utf-8"))
        _ini_bytes.extend(
            bytes(
                "\n".join(
                    [
                        f'{cfg}="{type_to_str(getattr(self, cfg))}"'
                        for cfg in get_not_ignored_slots(self)
                        if cfg != "section_name"
                    ]
                ),
                "utf-8",
            )
        )
        _ini_bytes.extend(b"\n\n")
        return _ini_bytes

    for cfg_name in get_not_ignored_slots(self):
        cfg = getattr(self, cfg_name)
        _ini_bytes.extend(_parse_config_list(cfg, cfg_name, title))

    _ini_bytes.pop()  # at least remove second line-break

    return _ini_bytes


def to_ini(self, out_cfg: Optional[str] = None) -> Optional[bytearray]:
    """Method to write an instance of the main config class of the module into an ini file."""
    return write_config(self.to_ini_bytes(self.__class__.__name__.lower()), out_cfg)


def create_ini_file_loader(
    cls: Config,
):
    """Method to create file loader for ini files."""

    def from_ini(*args, **kwargs):
        return create_file_loader(cls=cls, loader=ini_load)(*args, **kwargs)  # type: ignore

    return from_ini


# def create_ini_group_loader(
#     cls=None,
# ) -> Callable[..., ConfigGroup]:
#     """Method to create file loader for ini files."""
#
#     def from_ini(*args, **kwargs) -> ConfigGroup:
#         return create_group_loader(cls=cls)(*args, **kwargs)  # type: ignore
#
#     return from_ini



import logging
from typing import Any, Callable, List, Optional, Union, cast

from cornflakes.decorator._types import Config, ConfigGroup, DataclassProtocol, Loader
from cornflakes.decorator.config._config_group import config_group
from cornflakes.decorator.config._helper import get_default_loader
from cornflakes.decorator.config.dict import create_dict_file_loader
from cornflakes.decorator.config.ini import create_ini_file_loader
from cornflakes.decorator.config.yaml import create_yaml_file_loader
from cornflakes.decorator.dataclass import dataclass


def config(
    config_cls=None,
    files: Optional[Union[List[str], str]] = None,
    sections: Optional[Union[List[str], str]] = None,
    use_regex: Optional[bool] = False,
    is_list: Optional[Union[bool, int]] = False,
    default_loader: Optional[Loader] = None,
    allow_empty: Optional[bool] = False,
    filter_function: Optional[Callable[..., bool]] = None,
    **kwargs,
) -> Union[Union[Config, ConfigGroup], Callable[..., Union[Config, ConfigGroup]]]:
    """Config decorator to parse Ini Files and implements config loader methods to config-classes.

    :param config_cls: Config class
    :param files: Default config files
    :param sections: Default config sections
    :param use_regex: Flag to eval all sections by regex
    :param is_list: Flag to load Config as List of this class
    :param default_loader: Default config parser method (enum)
    :param kwargs: Default configs to overwrite dataclass args
    :param allow_empty: Flag that allows empty config result
    :param filter_function: Optional filter method for config
    :param kwargs: Additional args for custom dataclass. (dict_factory, eval_env, validate. ...).

    :returns: wrapped class or the wrapper itself with the custom default arguments if the config class is not
    """
    sections = sections if isinstance(sections, list) else [sections] if sections else []
    files = files if isinstance(files, list) else [files] if files else []
    if not default_loader:
        default_loader = get_default_loader(files)

    def wrapper(cls):
        """Wrapper function for the config decorator config_decorator."""
        # Check __annotations__
        if not hasattr(cls, "__annotations__"):
            logging.warning(f"{cls.__name__} missing annotations for values!")
            return cls

        # Check is config
        if any([hasattr(slot, "__config_sections__") for slot in cls.__annotations__.values()]):
            logging.warning(
                "Wrapper config not working for a subset of config classes. "
                f"Please use {config_group.__name__} instead."
            )
            return config_group(
                config_cls=cls,
                files=files,
                **kwargs,
            )(cls)

        cls: Union[DataclassProtocol, Any] = dataclass(cls, **kwargs)
        cls.__config_sections__ = sections
        cls.__config_files__ = files
        cls.__multi_config__ = use_regex
        cls.__config_list__ = is_list
        cls.__allow_empty_config__ = allow_empty
        cls.__config_filter_function__ = filter_function
        cls.from_yaml = staticmethod(create_yaml_file_loader(cls=cls))
        cls.from_ini = staticmethod(create_ini_file_loader(cls=cls))  # class not dependent method
        cls.from_dict = staticmethod(create_dict_file_loader(cls=cls))
        cls.from_file = getattr(cls, str(default_loader.value), cls.from_dict)

        return cast(Config, cls)

    if config_cls:
        return wrapper(config_cls)  # type: ignore
    return wrapper



import collections.abc
from os import environ
from os.path import abspath, expanduser
from typing import Any, Dict, List, Optional, Type, Union, cast

import yaml
from yaml import SafeLoader, UnsafeLoader


def _update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = _update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def _get_updated_key(config: Dict, values: List[str], defaults: Dict):
    result = None
    for value in values:
        key = config.get(value, defaults.get(value, None))
        if key:
            result = key
    return result


def _get_values(config, value: Union[str, List[str]], defaults: Dict):
    return (
        config
        if isinstance(config, str)
        else _get_updated_key(config, value, defaults)
        if isinstance(value, list)
        else config.get(value, defaults.get(value))
    )


def _get_section(
    config: dict,
    keys: dict,
    defaults: dict,
) -> dict:
    return {sub_key: _get_values(config, sub_value, defaults) for sub_key, sub_value in keys.items()}


def _get_all_sections(
    config: dict,
    sections: dict,
    keys: dict,
    defaults: dict,
) -> dict:
    new_config: dict = {}
    for section_key, section_names in sections.items():
        if not section_key:
            if isinstance(section_names, str):
                _update(new_config, _get_section(config.get(section_names, {}), keys, defaults))
                continue
            for section in section_names:
                _update(new_config, _get_section(config.get(section, {}), keys, defaults))
        if isinstance(section_names, str):
            _update(new_config, {section_key: _get_section(config.get(section_names, {}), keys, defaults)})
            continue
        for section in section_names:
            _update(new_config, {section_key: _get_section(config.get(section, {}), keys, defaults)})
    return new_config


def _yaml_read(
    file: str,
    sections: dict,
    keys: dict,
    defaults: dict,
    loader: Optional[Union[Type[UnsafeLoader], Type[SafeLoader]]],
) -> dict:
    with open(abspath(expanduser(file)), "rb") as f:
        data = f.read()
        if not loader:
            loader = yaml.UnsafeLoader if b": !!" in data else yaml.SafeLoader
        config = yaml.load(data, Loader=loader)  # noqa: S506
        if not sections:
            if not keys:
                return config
            if not defaults:
                defaults = keys
            return _get_all_sections(config, _to_map(list(config.keys())), keys, defaults)
        return _get_all_sections(config, sections, keys, defaults)


def _to_map(obj: Optional[Union[dict, list, tuple, str]]) -> dict:
    return (
        isinstance(obj, str)
        and {obj: obj}  # string
        or (isinstance(obj, list) or isinstance(obj, tuple))
        and {key: key for key in obj}  # list
        or cast(dict, obj)
        or {}
    )  # dict


def yaml_load(
    files: Union[str, List[str], Dict[Optional[str], Union[str, List[str]]]],
    sections: Optional[Union[str, List[str], Dict[Optional[str], Union[str, List[str]]]]] = None,
    keys: Optional[Union[str, List[str], Dict[Optional[str], Union[str, List[str]]]]] = None,
    defaults: Optional[Union[str, List[str], Dict[str, Any]]] = None,
    eval_env: bool = False,
    loader: Optional[Union[Type[SafeLoader], Type[UnsafeLoader]]] = None,
):
    """Yaml Wrapper for reading yaml files in a generic way."""
    files = _to_map(files)
    sections = _to_map(sections)
    keys = _to_map(keys)
    defaults = _to_map(defaults)

    if eval_env:
        defaults.update(
            {
                key: var
                for key, value in keys.items()
                if (
                    var := (
                        env_var
                        if isinstance(value, str) and (env_var := environ.get(value))
                        else env_vars[-1]
                        if (
                            env_vars := [
                                env_var
                                for sub_value in value
                                if isinstance(value, list) and (env_var := environ.get(sub_value))
                            ]
                        )
                        else None
                    )
                )
            }
        )

    config: dict = {}
    for file_key, file_names in files.items():
        if not file_key:
            if isinstance(file_names, str):
                _update(config, _yaml_read(file_names, sections, keys, defaults, loader))
                return config

            for file in file_names:
                _update(config, _yaml_read(file, sections, keys, defaults, loader))
                return config

        if isinstance(file_names, str):
            _update(config, {file_key: _yaml_read(file_names, sections, keys, defaults, loader)})
            return config

        for file in file_names:
            _update(config, {file_key: _yaml_read(file, sections, keys, defaults, loader)})

    return config



from dataclasses import Field as DataclassField
from dataclasses import MISSING
from typing import Any, Callable, List, Optional, Union

from cornflakes.decorator._types import WITHOUT_DEFAULT, _WithoutDefault

_MISSING_TYPE = type(MISSING)


class Field(DataclassField):
    """Instances of dataclasses Field wrapped for configs."""

    __slots__ = (*getattr(DataclassField, "__slots__", ()), "validator", "ignore", "alias")

    def __init__(
        self,
        default: Union[_MISSING_TYPE, Any] = MISSING,
        default_factory: Union[_MISSING_TYPE, _WithoutDefault, Callable[..., Any]] = MISSING,
        init: Optional[bool] = True,
        repr: Optional[bool] = True,
        hash: Optional[Union[bool, _MISSING_TYPE]] = None,
        compare: Optional[bool] = True,
        metadata: Optional[bool] = None,
        kw_only: Union[_MISSING_TYPE, bool] = MISSING,
        validator: Optional[Union[Callable[[str], Any], _MISSING_TYPE]] = MISSING,
        alias: Optional[Union[List[str], str]] = None,
        ignore: Optional[bool] = False,
    ):
        """Init Field method.

        :param default: Field default
        :param default_factory: Field factory
        :param init: Field
        :param repr: Field
        :param hash: Field
        :param compare: Field
        :param metadata: Field
        :param kw_only: Field
        :param validator: Field
        :param ignore: Field will be ignored when writing to_ini etc.
        :param alias: Aliases for config data (will be used to read from config files).
        """
        super().__init__(
            **{
                key: value
                for key, value in (
                    ("default", default),  # type: ignore
                    ("default_factory", default_factory),
                    ("init", init),
                    ("repr", repr),
                    ("hash", hash),
                    ("compare", compare),
                    ("metadata", metadata),
                    ("kw_only", kw_only),
                )
                if key in DataclassField.__init__.__code__.co_varnames
            }
        )
        self.ignore = ignore
        self.validator = validator
        self.alias = alias

    def __repr__(self):
        """Repr Field method."""
        return (
            f"{DataclassField.__repr__(self)[:-1]},"
            f"validator={self.validator}, "
            f"ignore={self.ignore}"
            f"alias={self.alias})"
        )


def field(
    *,
    default: Union[_MISSING_TYPE, Any] = MISSING,
    default_factory: Union[_MISSING_TYPE, Callable] = MISSING,
    init: Optional[bool] = True,
    repr: Optional[bool] = True,
    hash: Optional[Union[bool, _MISSING_TYPE]] = None,
    compare: Optional[bool] = True,
    metadata: Optional[bool] = None,
    kw_only: Union[_MISSING_TYPE, bool] = MISSING,
    validator: Optional[Union[Callable[[str], Any], _MISSING_TYPE]] = MISSING,
    ignore: Optional[bool] = False,
    alias: Optional[Union[List[str], str]] = None,
    no_default: Optional[bool] = None,
):
    """This function is used instead of exposing Field creation directly.

    So that a type checker can be told (via overloads) that this is a function whose type depends on its parameters.
    This function is used instead of exposing Field creation directly,
    so that a type checker can be told (via overloads) that this is a
    function whose type depends on its parameters.
    """
    if default is not MISSING and default_factory is not MISSING:
        raise ValueError("cannot specify both default and default_factory")

    if default is MISSING and default_factory is MISSING and no_default:
        default_factory = WITHOUT_DEFAULT

    new_field = Field(default, default_factory, init, repr, hash, compare, metadata, kw_only, validator, alias, ignore)
    return new_field



"""Default Global Click Options."""
from cornflakes.common import patch_module
from cornflakes.click.options._bg_process import bg_process_option
from cornflakes.click.options._verbose import verbose_option
from cornflakes.click.options._global import global_option
from cornflakes.click.options._auto_option import auto_option

__all__ = ["verbose_option", "bg_process_option", "global_option", "auto_option"]

patch_module(globals())



from contextlib import suppress
import inspect
from itertools import chain
from os import environ
from typing import Any, Callable, Optional, Union, get_args

from _cornflakes import eval_type
from cornflakes.common import extract_var_names, wrap_kwargs
from cornflakes.decorator._types import WITHOUT_DEFAULT, Config, ConfigGroup, DataclassProtocol

SpecialForm = type(Optional)

_empty = getattr(inspect, "_empty", None)


def enforce_types(config: Union[DataclassProtocol, Config, ConfigGroup], validate=False):  # noqa: C901
    """Adds a simple decorator enforce_types that enables enforcing strict typing on a function or dataclass using annotations.

    Works with collection types and subtypes for example Dict[str, Tuple[int, int]], and with special types as Optional and Any.
    Base Reference: https://github.com/matchawine/python-enforce-typing
    """
    validators = {
        key: validator
        for key, value in config.__dataclass_fields__.items()
        if callable(validator := getattr(value, "validator", key))
    }

    required_keys = [
        key
        for key, value in config.__dataclass_fields__.items()
        if getattr(value, "default_factory", None) == WITHOUT_DEFAULT
    ]

    def _check_type(type_hint: Any, key, value):
        if isinstance(type_hint, type(None)):
            return

        if isinstance(type_hint, SpecialForm):
            return value

        with suppress(KeyError):

            actual_type: Any = getattr(type_hint, "__origin__", getattr(type_hint, "type", type_hint))

            if isinstance(actual_type, SpecialForm):
                actual_type = getattr(type_hint, "__args__", type_hint)

            if isinstance(actual_type, list) or isinstance(actual_type, tuple):
                actual_types = [t for t in actual_type if t is not None]
                return next(item for item in [_check_type(t, key, value) for t in actual_types] if item is not None)

            if actual_type in [list, tuple]:
                if not isinstance(value, list) and not isinstance(value, tuple):
                    raise TypeError(
                        f"Expected type '{type_hint}' for attribute '{key}' but received type '{type(value)}')"
                    )
                actual_types = [t for t in get_args(type_hint) if t is not None] or [str]
                return actual_type(chain([_check_type(t, key, val) for val in value for t in actual_types]))

            if not isinstance(value, actual_type):
                try:
                    if not validate:
                        raise TypeError(
                            f"Expected type '{type_hint}' for attribute '{key}' but received type '{type(value)}')"
                        )
                    return actual_type(value)  # type: ignore
                except Exception as exc:
                    raise Exception(
                        f"Expected type '{type_hint}' for attribute '{key}' but received type '{type(value)}')"
                    ) from exc

            return value

    def _get_env_vars():
        return {key: eval_type(environ[key]) for key in config.__dataclass_fields__.keys() if key in environ.keys()}

    def _validate(self, values, key, callback: Callable[..., Any]):
        try:
            co_varnames = extract_var_names(callback)
            kwargs = {}
            kwargs.update(co_varnames)
            kwargs.update({"self": self, "values": values, "key": key})
            kwargs.update(values)
            kwargs = {key: value for key, value in kwargs.items() if key in co_varnames.keys() and value is not _empty}
            if len(missing := [key for key in co_varnames.keys() if key not in kwargs.keys()]) > 1:
                raise TypeError(f"Argument not provided: {missing})")
            if len(missing) == 1:
                return callback(values.pop(key), **kwargs)
            return callback(**kwargs)
        except Exception as exc:
            raise Exception(f"Failed to validate {key} with {callback}!") from exc

    def _process_validator(self, values, validators: dict, **kwargs):
        return {
            key: _validate(self, values, key, callback) for key, callback in validators.items() if key in kwargs.keys()
        }

    def _process_type_checking(**kwargs):
        return {
            key: _check_type(getattr(config, "__annotations__", {})[key], key, value)
            for key, value in kwargs.items()
            if key in config.__dataclass_fields__.keys()
        }

    def decorate(func):
        @wrap_kwargs(func)
        def wrapper(self, *args, **kwargs):
            kwargs.update(dict(zip(func.__code__.co_varnames[1:], args)))
            default_kwargs = {}
            default_kwargs.update(kwargs)

            if self.__eval_env__:
                default_kwargs.update(_get_env_vars())

            default_kwargs.update(_process_validator(self, default_kwargs, validators, **default_kwargs))

            default_kwargs.update(_process_type_checking(**default_kwargs))
            default_kwargs.pop("self", None)
            if missing_keys := [key for key in required_keys if key not in default_kwargs.keys()]:
                raise ValueError(f"Missing required values for keys {missing_keys}")

            return func(self, **default_kwargs)  # type: ignore

        return wrapper

    config.__init__ = decorate(config.__init__)

    return config



from dataclasses import field
from os import getenv
from typing import Dict, List, Literal, Optional, Protocol, Union

from click import Option

from cornflakes.decorator.config import config


class GlobalOption(Protocol):
    """GlobalOption Protocol which requires params."""

    params: List[Option]


@config
class RichConfig:
    """DataClass for click config-values."""

    # Default styles
    STYLE_OPTION: str = "bold cyan"
    STYLE_ARGUMENT: str = "bold cyan"
    STYLE_SWITCH: str = "bold green"
    STYLE_METAVAR: str = "bold yellow"
    STYLE_METAVAR_APPEND: str = "dim yellow"
    STYLE_METAVAR_SEPARATOR: str = "dim"
    STYLE_HEADER_TEXT: str = ""
    STYLE_FOOTER_TEXT: str = ""
    STYLE_USAGE: str = "yellow"
    STYLE_USAGE_COMMAND: str = "bold"
    STYLE_DEPRECATED: str = "red"
    STYLE_HELPTEXT_FIRST_LINE: str = ""
    STYLE_HELPTEXT: str = "dim"
    STYLE_OPTION_HELP: str = ""
    STYLE_OPTION_DEFAULT: str = "dim"
    STYLE_OPTION_ENVVAR: str = "dim yellow"
    STYLE_REQUIRED_SHORT: str = "red"
    STYLE_REQUIRED_LONG: str = "dim red"
    STYLE_OPTIONS_PANEL_BORDER: str = "dim"
    ALIGN_OPTIONS_PANEL: str = "left"
    STYLE_OPTIONS_TABLE_SHOW_LINES = False
    STYLE_OPTIONS_TABLE_LEADING = 0
    STYLE_OPTIONS_TABLE_PAD_EDGE = False
    STYLE_OPTIONS_TABLE_PADDING = (0, 1)
    STYLE_OPTIONS_TABLE_BOX: str = ""
    STYLE_OPTIONS_TABLE_ROW_STYLES = None
    STYLE_OPTIONS_TABLE_BORDER_STYLE = None
    STYLE_COMMANDS_PANEL_BORDER: str = "dim"
    ALIGN_COMMANDS_PANEL: str = "left"
    STYLE_COMMANDS_TABLE_SHOW_LINES = False
    STYLE_COMMANDS_TABLE_LEADING = 0
    STYLE_COMMANDS_TABLE_PAD_EDGE = False
    STYLE_COMMANDS_TABLE_PADDING = (0, 1)
    STYLE_COMMANDS_TABLE_BOX: str = ""
    STYLE_COMMANDS_TABLE_ROW_STYLES = None
    STYLE_COMMANDS_TABLE_BORDER_STYLE = None
    STYLE_ERRORS_PANEL_BORDER: str = "red"
    ALIGN_ERRORS_PANEL: str = "left"
    STYLE_ERRORS_SUGGESTION: str = "dim"
    STYLE_ABORTED: str = "red"
    MAX_WIDTH = int(getenv("TERMINAL_WIDTH")) if getenv("TERMINAL_WIDTH") else None  # type: ignore
    COLOR_SYSTEM: Optional[Literal["auto", "standard", "256", "truecolor", "windows"]] = "auto"
    # Set to None to disable colors
    FORCE_TERMINAL = True if getenv("GITHUB_ACTIONS") or getenv("FORCE_COLOR") or getenv("PY_COLORS") else None

    # Fixed strings
    HEADER_TEXT: Optional[str] = None
    HEADER_LOGO: Optional[str] = None
    FOOTER_TEXT: Optional[str] = None
    DEPRECATED_STRING: str = "(Deprecated) "
    DEFAULT_STRING: str = "[default: {}]"
    ENVVAR_STRING: str = "[env var: {}]"
    REQUIRED_SHORT_STRING: str = "*"
    REQUIRED_LONG_STRING: str = "[required]"
    RANGE_STRING: str = " [{}]"
    APPEND_METAVARS_HELP_STRING: str = "({})"
    ARGUMENTS_PANEL_TITLE: str = "Arguments"
    OPTIONS_PANEL_TITLE: str = "Options"
    COMMANDS_PANEL_TITLE: str = "Commands"
    ERRORS_PANEL_TITLE: str = "Error"
    ERRORS_SUGGESTION: Optional[str] = None  # Default: Try 'cmd -h' for help. Set too False to disable.
    ERRORS_EPILOGUE: Optional[str] = None
    ABORTED_TEXT: str = "Aborted."

    # Behaviours
    SHOW_ARGUMENTS: bool = True  # Show positional arguments
    SHOW_METAVARS_COLUMN: bool = True  # Show a column with the option metavar (eg. INTEGER)
    APPEND_METAVARS_HELP: bool = False  # Append metavar (eg. [TEXT]) after the help text
    APPEND_METAVARS_REQUIRED: bool = False  # Append metavar REQUIRED_LONG_STRING (eg. [required]) after the help text
    GROUP_ARGUMENTS_OPTIONS: bool = False  # Show arguments with options instead of in own panel
    OPTION_ENVVAR_FIRST: bool = False  # Show env vars before option help text instead of avert
    USE_RST: bool = True
    USE_MARKDOWN: bool = False  # Parse help strings as markdown
    USE_MARKDOWN_EMOJI: bool = True  # Parse emoji codes in markdown :smile:
    USE_RICH_MARKUP: bool = False  # Parse help strings for rich markup (eg. [red]my text[/])
    # Define sorted groups of panels to display subcommands
    USE_CLICK_SHORT_HELP: bool = False  # Use click's default function to truncate help text

    VERSION_INFO: bool = False

    CONTEXT_SETTINGS: Dict[str, List[Dict[str, Union[str, List[str]]]]] = field(default_factory=dict)

    COMMAND_GROUPS: Dict[str, List[Dict[str, Union[str, List[str]]]]] = field(default_factory=dict)
    # Define sorted groups of panels to display options and arguments
    OPTION_GROUPS: Dict[str, List[Dict[str, Union[str, List[str]]]]] = field(default_factory=dict)
    # Add basic global options (verbose)
    GLOBAL_OPTIONS: List[GlobalOption] = field(default_factory=list)



"""cornflakes config decorator module."""
from cornflakes.common import patch_module
from cornflakes.decorator.config._config import config, Config
from cornflakes.decorator.config._config_group import config_group, ConfigGroup
from cornflakes.decorator.config._helper import is_config
from cornflakes.decorator.config._helper import config_files

__all__ = ["config", "config_group", "is_config", "Config", "ConfigGroup", "config_files"]

patch_module(globals())



from functools import wraps
from inspect import signature
from typing import Callable, Optional, Union

from click import get_current_context
from click.core import Context

from cornflakes.click.rich._rich_argument import RichArg
from cornflakes.click.rich._rich_command import RichCommand
from cornflakes.click.rich._rich_group import RichGroup

F = Callable[..., Union[RichCommand, RichGroup, RichArg]]


def rich_global_option_wrapper(click_func, *wrap_args, pass_context: Optional[bool] = None, **wrap_kwargs) -> F:
    """Wrapper Method for rich command / group."""

    def global_option_click_decorator(func):

        click_cls = click_func(*wrap_args, **wrap_kwargs)(func)

        @wraps(func)
        def click_callback(*args, **kwargs):
            kwargs["self"]: RichGroup = func
            kwargs["parent"]: RichGroup = click_cls
            if pass_context:
                kwargs["ctx"]: Optional["Context"] = get_current_context()
            if click_cls.config:
                if click_cls.config.GLOBAL_OPTIONS and func.__module__ != "cornflakes.click":
                    for option_obj in click_cls.config.GLOBAL_OPTIONS:
                        option_obj(
                            *args,
                            **{
                                key: value
                                for key, value in kwargs.items()
                                if key in signature(option_obj).parameters.keys()
                            },
                        )
            return func(
                *args, **{key: value for key, value in kwargs.items() if key in signature(func).parameters.keys()}
            )

        click_cls.callback = click_callback
        return click_cls

    return global_option_click_decorator



from dataclasses import astuple, fields, is_dataclass
from typing import Any

from cornflakes.decorator.config._helper import tuple_factory


def to_tuple(self) -> Any:
    """Method to convert Dataclass with slots to dict."""
    if not is_dataclass(self):
        return self
    new_tuple = astuple(self, tuple_factory=tuple_factory(self))
    if not (
        isinstance(new_tuple, (list, tuple))
        or any([is_dataclass(f.type) or f.default_factory == list or isinstance(f.default, list) for f in dc_fields])
        if (dc_fields := fields(self))
        else True
    ):
        return new_tuple
    if isinstance(new_tuple, tuple):
        new_tuple = list(new_tuple)
    for idx, f in enumerate(dc_fields):
        if is_dataclass(value := getattr(self, f.name)):
            new_tuple[idx] = value.to_tuple()
        if isinstance(value, list):
            for sub_idx, sub_value in enumerate(value):
                if is_dataclass(sub_value):
                    value[sub_idx] = sub_value.to_tuple()
            new_tuple[idx] = value
    if isinstance(new_tuple, list):
        new_tuple = tuple(new_tuple)  # cast to tuple
    return new_tuple



import dataclasses


def add_slots(cls):
    """Need to create a new class, since we can't set __slots__ after a class has been created.

    This decorator is needed to support lower python versions (newer versions has slots function inside dataclasses itself).
    """
    # Make sure __slots__ isn't already set.
    if "__slots__" in cls.__dict__:
        raise TypeError(f"{cls.__name__} already specifies __slots__")

    # Create a new dict for our new class.
    cls_dict = dict(cls.__dict__)
    field_names = tuple(f.name for f in dataclasses.fields(cls))
    cls_dict["__slots__"] = field_names
    for field_name in field_names:
        # Remove our attributes, if present. They'll still be
        #  available in _MARKER.
        cls_dict.pop(field_name, None)
    # Remove __dict__ itself.
    cls_dict.pop("__dict__", None)
    # And finally create the class.
    qualname = getattr(cls, "__qualname__", None)
    cls = type(cls)(cls.__name__, cls.__bases__, cls_dict)
    if qualname is not None:
        cls.__qualname__ = qualname
    return cls



from typing import Callable, Dict, List, Optional, Type, Union

import yaml
from yaml import SafeLoader, UnsafeLoader

from cornflakes.decorator._types import Config
from cornflakes.decorator.config._load_config import create_file_loader
from cornflakes.decorator.config._write_config import write_config
from cornflakes.parser import yaml_load


def specific_yaml_loader(loader: Union[Type[SafeLoader], Type[UnsafeLoader]] = SafeLoader):
    """Wrapper method to predefine yaml loader parameter."""

    def _yaml_loader(*args, **kwargs):
        return yaml_load(*args, loader=loader, **kwargs)

    return _yaml_loader


def to_yaml_bytes(self, *args, **kwargs) -> Optional[bytearray]:
    """Method to write an instance of the main config class of the module into a yaml bytearray."""
    return yaml.dump({self.__class__.__name__.lower(): self.to_dict()}, *args, **kwargs).encode("utf-8")


def to_yaml(self, out_cfg: Optional[str] = None, *args, **kwargs) -> Optional[bytearray]:
    """Method to write an instance of the main config class of the module into a yaml file."""
    return write_config(self.to_yaml_bytes(*args, **kwargs), out_cfg)


def create_yaml_file_loader(
    cls: Config,
) -> Callable[..., Dict[str, Optional[Union[Config, List[Config]]]]]:
    """Method to create file loader for yaml files."""

    def from_yaml(
        *args, loader: Union[Type[SafeLoader], Type[UnsafeLoader]] = SafeLoader, **kwargs
    ) -> Dict[str, Optional[Union[Config, List[Config]]]]:
        _from_yaml = create_file_loader(cls=cls, loader=specific_yaml_loader(loader=loader))  # type: ignore
        return _from_yaml(*args, **kwargs)

    return from_yaml


# def create_yaml_group_loader(
#     cls=None,
# ) -> Callable[..., ConfigGroup]:
#     """Method to create file loader for yaml files."""
#
#     def from_yaml(*args, loader: Union[Type[SafeLoader], Type[UnsafeLoader]] = SafeLoader, **kwargs) -> ConfigGroup:
#         _from_yaml = create_group_loader(cls=cls, loader=specific_yaml_loader(loader=loader))  # type: ignore
#         return _from_yaml(*args, **kwargs)
#
#     return from_yaml



from dataclasses import InitVar, field
from typing import Optional
from urllib.parse import ParseResult, urlparse, urlunparse

from cornflakes.decorator.config.tuple import to_tuple
from cornflakes.decorator.dataclass._dataclass import dataclass as data


@data(
    slots=True,
    frozen=False,
    tuple_factory=lambda self, x: urlunparse(x[:6]),  # unparse to string
    dict_factory=lambda self, x: str(self),  # to string -> tuple -> unparse to string
)
class AnyUrl:
    """Database URL.

    :cvar url: URL to init the whole object (will be overwritten with other args).
    :cvar scheme: The scheme of the url
    :cvar netloc: user / pw / host and port of the url
    :cvar path: path of the url
    :cvar params: url parameters
    :cvar query: url query
    :cvar fragment: url fragment
    :cvar hostname: url hostname (overwrites the netloc)
    :cvar port: url port (overwrites the netloc)
    :cvar username: url username (overwrites the netloc)
    :cvar password: url password (overwrites the netloc)
    """

    url: InitVar[Optional[str]] = None
    scheme: Optional[str] = field(default=None, init=True)
    netloc: Optional[str] = field(default=None, init=True)
    path: Optional[str] = field(default=None, init=True)
    params: Optional[str] = field(default=None, init=True)
    query: Optional[str] = field(default=None, init=True)
    fragment: Optional[str] = field(default=None, init=True)
    hostname: Optional[str] = field(default=None, init=True, repr=False)
    port: Optional[int] = field(default=None, init=True, repr=False)
    username: Optional[str] = field(default=None, init=True, repr=False)
    password: Optional[str] = field(default=None, init=True, repr=False)

    def __init_parsed(self, parsed: ParseResult):
        for slot in getattr(self, "__dataclass_fields__", {}).values():
            if not getattr(self, slot.name, None) and not isinstance(slot.type, InitVar):
                setattr(self, slot.name, getattr(parsed, slot.name, None))

    def __post_init__(self, url: Optional[str] = None) -> None:
        """Post init."""
        if url:
            parsed = urlparse(url)
            self.__init_parsed(parsed)
        else:
            if not self.netloc or self.username or self.password or self.hostname or self.port:
                login = f"{self.username}:{self.password}@" if self.username else ""
                parsed = urlparse(str(self))
                port = f":{self.port or parsed.port}" if self.port or parsed.port else ""
                hostname = self.hostname if self.hostname else parsed.hostname
                self.netloc = f"{login}{hostname}{port}"
            parsed = urlparse(to_tuple(self))
            self.__init_parsed(parsed)

    def __str__(self) -> str:
        """Any url string."""
        return to_tuple(self)



from os import path
from typing import Optional, Union


def write_config(config_bytes: bytearray, out_cfg: Optional[str] = None) -> Union[bytearray, None]:
    """Method to write bytes into file."""
    if not out_cfg:
        return config_bytes
    with open(path.abspath(path.expanduser(out_cfg)), "wb") as f:
        f.write(config_bytes)
    return None



from typing import Any, Callable, List, Optional, Union, cast

from cornflakes.decorator._types import ConfigGroup, DataclassProtocol
from cornflakes.decorator.config._load_config_group import create_group_loader
from cornflakes.decorator.dataclass import dataclass


def config_group(
    config_cls=None,
    files: Optional[Union[str, List[str]]] = None,
    allow_empty: Optional[bool] = None,
    filter_function: Optional[Callable[..., bool]] = None,
    **kwargs,
) -> Union[ConfigGroup, Callable[..., ConfigGroup]]:
    """Config decorator with a Subset of configs to parse Ini Files.

    :param config_cls: Config class
    :param files: Default config files
    :param allow_empty: Flag that allows empty config result
    :param filter_function: Optional filter method for config
    :param kwargs: Additional args for custom dataclass. (dict_factory, eval_env. ...).

    :returns: wrapped class or the wrapper itself with the custom default arguments if the config class is not

    """
    files = files if isinstance(files, list) else [files] if files else []

    kwargs.pop("validate", None)  # no validation for group

    def wrapper(cls) -> ConfigGroup:
        cls: Union[DataclassProtocol, Any] = dataclass(cls, **kwargs)
        cls.__config_files__ = files
        cls.__allow_empty_config__ = allow_empty
        cls.__config_filter_function__ = filter_function

        cls.from_file = staticmethod(create_group_loader(cls=cls))

        return cast(ConfigGroup, cls)

    if config_cls:
        return wrapper(config_cls)
    return wrapper



# from collections import OrderedDict
import logging
from typing import Any, Callable, Dict, List, Optional, Union

from cornflakes.decorator._types import Config, ConfigGroup
from cornflakes.decorator.config._helper import is_config, is_config_list


def create_group_loader(cls) -> Callable[..., ConfigGroup]:
    """Config decorator to parse Ini Files and implements from_file method to config-group-classes.

    :param cls: Config class

    :returns: wrapped class or the wrapper itself with the custom default arguments if the config group class is not
    """

    def from_file(
        files: Optional[Union[Dict[str, Union[List[str], str]], List[str], str]] = None,
        sections: Optional[Union[List[str], str]] = None,
        config_dict: Optional[Dict[str, Any]] = None,
        filter_function: Optional[Callable[[Config], bool]] = None,
        eval_env: Optional[bool] = None,
        allow_empty: Optional[bool] = None,
        *slot_args,
        **slot_kwargs,
    ) -> ConfigGroup:
        """Config parser from config files.

        :param files: Default config files
        :param sections: Default config sections
        :param config_dict: Config dictionary to pass already loaded configs
        :param slot_kwargs: Default configs to overwrite passed class
        :param filter_function: Optional filter method for config
        :param eval_env: Flag to evaluate environment variables into default values.
        :param allow_empty: Flag that allows empty config result -> e.g. emtpy list
        :param slot_args: Default configs to overwrite passed class
        :param slot_kwargs: Default configs to overwrite passed class

        :returns: Nested Lists of Config Classes

        """
        if not files:
            files = cls.__config_files__

        # if not config_dict:
        #     config_dict = OrderedDict(loader({None: files}, eval_env=eval_env))
        #     print(config_dict)
        # logging.debug(f"Read config with sections: {config_dict.keys()}")
        for slot_class in list(getattr(cls, "__annotations__", {}).values())[len(slot_args) :]:
            if is_config_list(slot_class):
                slot_class = slot_class.__args__[0]
            if is_config(slot_class):
                slot_kwargs.update(
                    slot_class.from_file(
                        files=files,
                        sections=sections,
                        config_dict=config_dict,
                        filter_function=filter_function or cls.__config_filter_function__,
                        eval_env=eval_env,
                        allow_empty=allow_empty or cls.__allow_empty_config__,
                    )
                )
        error_args = [key for key in slot_kwargs if key not in cls.__dataclass_fields__]
        if error_args:
            logging.warning(f"The variables {error_args} in **{cls.__name__}** are not defined!")
            logging.warning("Use generate_group in build script to auto generate the config group!")

        return cls(*slot_args, **{key: value for key, value in slot_kwargs.items() if key not in error_args})

    return from_file  # class not dependent method



"""cornflakes common module.
__________________________________
"""  # noqa: RST303 D205
from cornflakes.common._extract_var_names import extract_var_names
from cornflakes.common._default_ca_path import default_ca_path
from cornflakes.common._type_to_str import type_to_str
from cornflakes.common._types import datetime_ms, unquoted_string
from cornflakes.common._patch_module import patch_module
from cornflakes.common._wrap_kwargs import wrap_kwargs

__all__ = [
    "default_ca_path",
    "type_to_str",
    "datetime_ms",
    "extract_var_names",
    "unquoted_string",
    "patch_module",
    "wrap_kwargs",
]

patch_module(globals())



from inspect import getfile
from typing import Any, Callable, Optional, Union

from click import BaseCommand, Group, style, version_option
import pkg_resources

from cornflakes.click import RichConfig, RichGroup, group
from cornflakes.decorator._types import Loader
from cornflakes.decorator.config import Config


def click_cli(  # noqa: C901
    callback: Optional[Callable] = None,
    config: Optional[Config] = None,
    files: Optional[str] = None,
    loader: Loader = Loader.DICT_LOADER,
    *args,
    **kwargs,
) -> Union[
    Callable[[Any], Callable[..., Union[BaseCommand, RichGroup]]], Callable[..., Union[BaseCommand, Group, RichGroup]]
]:
    """Function that creates generic click CLI Object."""
    if not config:
        if not files:
            config = RichConfig(*args, **kwargs)
        elif loader in [Loader.INI_LOADER, Loader.YAML_LOADER]:
            config = getattr(RichConfig, str(loader.name))(*args, **kwargs).popitem()[1]
        elif ".ini" in files:
            config = RichConfig.from_ini(files, *args, **kwargs).popitem()[1]
        elif ".yaml" in files:
            config = RichConfig.from_yaml(files, *args, **kwargs).popitem()[1]
        else:
            config = RichConfig(*args, **kwargs)

    def cli_wrapper(w_callback: Callable) -> Callable[..., Union[BaseCommand, Group, RichGroup]]:
        if not callable(w_callback):
            return w_callback

        module = getfile(w_callback)
        cli_group: Union[BaseCommand, Group, RichGroup] = group(module.split(".", 1)[0], config=config)(w_callback)
        if config.VERSION_INFO:
            name = w_callback.__qualname__
            __version = "0.0.1"

            if hasattr(w_callback, "__module__"):
                module = w_callback.__module__.split(".", 1)[0]
                if module != "__main__":
                    __version = pkg_resources.get_distribution(module).version

            version_args = {
                "prog_name": name,
                "version": __version,
                "message": style(
                    f"\033[95m{module}\033" f"[0m \033[95m" f"Version\033[0m: \033[1m" f"{__version}\033[0m"
                ),
            }
            cli_group = version_option(**version_args)(cli_group)

        if cli_group.config.GLOBAL_OPTIONS:
            for option_obj in cli_group.config.GLOBAL_OPTIONS:
                cli_group.params.extend(option_obj.params)
        if config.CONTEXT_SETTINGS:
            cli_group.context_settings = config.CONTEXT_SETTINGS
        return cli_group

    if callback:
        return cli_wrapper(callback)
    return cli_wrapper



from inspect import isclass
import logging
from typing import Any, Callable, Dict, List, Optional, Union

from click import Command, Group, option

from cornflakes.click.rich import RichCommand, RichGroup, argument
from cornflakes.decorator.config import Config, ConfigGroup, is_config

F = Callable[[Union[Command, Group, Callable[..., Any]]], Union[Command, Group, Callable[..., Any], Callable]]


def auto_option(config, **options) -> F:  # noqa: C901
    """Click Option Decorator to define a global option for cli decorator."""
    if not isclass(config):
        raise TypeError("config should be a class!")

    _is_config = is_config(config)

    def auto_option_decorator(callback: Union[Union[Command, Group, RichCommand, RichGroup], Callable[..., None]]):

        if not callable(callback):
            raise TypeError("Wrapped object should be a function!")

        def wrapper(config_args: Optional[tuple] = None, *args, **kwargs):
            config_args = (
                (None, None) if not config_args else (config_args[0], None) if len(config_args) == 1 else config_args
            )
            __config: Union[Dict[str, Union[Config, List[Config]]], ConfigGroup] = config.from_file(
                files=config_args[0], sections=config_args[1]
            )
            if not __config:
                logging.error(f"Config is empty {__config} for file {config_args[0]} and section {config_args[1]}")
                raise ValueError
            if _is_config:
                __config = __config.popitem()[1]
                if isinstance(__config, list):
                    __config = __config[1]
            kwargs.update(
                {
                    "config": __config,
                    "config_args": config_args,
                }
            )
            return callback(
                *args, **{key: value for key, value in kwargs.items() if key in callback.__code__.co_varnames}
            )

        if hasattr(callback, "params"):
            wrapper.params = callback.params

        configs = {}
        for line in config.__doc__.split("\n"):
            line = line.strip()
            if line[:5] == ":cvar":
                line = line[6:].split(":")
                configs.update({line[0]: line[1].strip()})
        configs.update(options)
        for slot_name in config.__dataclass_fields__.keys():
            wrapper = option(f"--{slot_name.replace('_', '-')}", help=configs.get(slot_name, ""))(wrapper)

        wrapper = argument("config_args", required=False, nargs=-1, help="Passed Config to Method")(wrapper)

        return wrapper

    return auto_option_decorator



import sys
from typing import Any, Dict, List, Optional, Union

from click import ClickException, Command, Context, Group, HelpFormatter, Parameter, exceptions

from cornflakes.click.rich._rich_click import get_rich_console, rich_abort_error, rich_format_error, rich_format_help
from cornflakes.click.rich._rich_command import RichCommand
from cornflakes.click.rich._rich_config import RichConfig as RichConfig


class RichGroup(Group):
    """Richly formatted click Group.

    Inherits click.Group and overrides help and error methods
    to print richly formatted output.
    """

    command_class = RichCommand
    group_class = type
    params: List[Parameter]
    name = ""
    context_settings: dict
    commands: Dict[str, Union[Command, RichCommand]]

    def callback(self):
        """Callback method with is wrapped over the command group."""

    def add_command(self, cmd: Union[Command, RichCommand], name: Optional[str] = None) -> None:
        """Registers another :class:`Command` with this group.

        If the name is not provided, the name of the command is used.
        """
        Group.add_command(self, cmd, name)

    def __init__(self, config: RichConfig = None, *args, **kwargs):
        """Init function of RichGroup with extra config argument."""
        if not config:
            config = RichConfig()
        super().__init__(*args, **kwargs)
        self.config = config
        self.console = get_rich_console(config=self.config)

    def __pass_config(self, config=None, console=None):
        if config:
            for _group in self.commands.values():
                if isinstance(_group, RichGroup) and _group:
                    _group.__pass_config(config, console)
                _group.config = config
                _group.console = console if console else get_rich_console(_group.config)
                if _group.config.GLOBAL_OPTIONS:
                    for option_obj in _group.config.GLOBAL_OPTIONS:
                        _group.params.extend(option_obj.params)

    def main(self, *args, standalone_mode: bool = True, **kwargs) -> Any:  # noqa: C901
        """Main function of RichGroup."""
        try:
            self.__pass_config(self.config, self.console)
            rv = super().main(*args, standalone_mode=False, **kwargs)  # type: ignore
            if not standalone_mode:
                return rv
        except ClickException as e:
            if not standalone_mode:
                raise
            rich_format_error(e, config=self.config, console=self.console)
            sys.exit(e.exit_code)
        except exceptions.Abort:
            if not standalone_mode:
                raise
            rich_abort_error(config=self.config, console=self.console)
            sys.exit(1)

    def format_help(self, ctx: Context, formatter: HelpFormatter) -> None:
        """Format function of RichGroup."""
        rich_format_help(self, ctx, formatter, config=self.config, console=self.console)



from typing import Callable

from click import Option


def global_option(*option_args, **option_kwargs) -> Callable:
    """Click Option Decorator to define a global option for cli decorator."""
    _option = Option(*option_args, **option_kwargs)

    def global_option_decorator(option_func):
        if not callable(option_func):
            raise TypeError("Wrapped object should be a function!")

        if not hasattr(option_func, "params"):
            option_func.params = []

        option_func.params.append(_option)

        return option_func

    return global_option_decorator



import inspect
import re
import typing
from typing import Any, Generator, List, Union

import click
from rich import box
from rich.align import Align
from rich.columns import Columns
from rich.console import Console, group
from rich.emoji import Emoji
from rich.highlighter import RegexHighlighter
import rich.markdown
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.theme import Theme
from rich_rst import RestructuredText

from cornflakes.click.rich._rich_config import RichConfig as RichConfig

# Support rich <= 10.6.0
# try:
# except ImportError:
# from rich.console import render_group as group

# Rich regex highlighter
class OptionHighlighter(RegexHighlighter):
    """Highlights our special options."""

    highlights = [
        r"(^|\W)(?P<switch>\-\w+)(?![a-zA-Z0-9])",
        r"(^|\W)(?P<option>\-\-[\w\-]+)(?![a-zA-Z0-9])",
        r"(^|\W)(?P<argument>[A-Z0-9\_]+)(?![_a-zA-Z0-9])",
        r"(?P<metavar>\<[^\>]+\>)",
        r"(?P<usage>Usage: )",
    ]


highlighter = OptionHighlighter()


def get_rich_console(config: RichConfig) -> Console:
    return Console(
        theme=Theme(
            {
                "option": config.STYLE_OPTION,
                "argument": config.STYLE_ARGUMENT,
                "switch": config.STYLE_SWITCH,
                "metavar": config.STYLE_METAVAR,
                "metavar_sep": config.STYLE_METAVAR_SEPARATOR,
                "usage": config.STYLE_USAGE,
            }
        ),
        highlighter=highlighter,
        color_system=config.COLOR_SYSTEM,
        force_terminal=config.FORCE_TERMINAL,
        width=config.MAX_WIDTH,
    )


def _make_rich_rext(
    text: str, config: RichConfig, style: str = ""
) -> Union[rich.markdown.Markdown, rich.text.Text, RestructuredText]:
    """Take a string, remove indentations, and return styled text.

    By default, return the text as a Rich Text with the request style.
    If USE_RICH_MARKUP is True, also parse the text for Rich markup strings.
    If USE_MARKDOWN is True, parse as Markdown.

    Only one of USE_MARKDOWN or USE_RICH_MARKUP can be True.
    If both are True, USE_MARKDOWN takes precedence.

    Args:
        text (str): Text to style
        style (str): Rich style to apply

    Returns:
        MarkdownElement or Text: Styled text object
    """
    # Remove indentations from input text
    text = inspect.cleandoc(text)
    if config.USE_RST:
        return RestructuredText(f"| {text}")
    if config.USE_MARKDOWN:
        if config.USE_MARKDOWN_EMOJI:
            text = Emoji.replace(text)
        return Markdown(text, style=style)
    if config.USE_RICH_MARKUP:
        return highlighter(Text.from_markup(text, style=style))
    else:
        return highlighter(Text(text, style=style))


@group()
def _get_help_text(obj: Union[click.Command, click.Group], config: RichConfig) -> Generator:
    """Build primary help text for a click command or group.

    Returns the prose help text for a command or group, rendered either as a
    Rich Text object or as Markdown.
    If the command is marked as depreciated, the depreciated string will be prepended.

    Args:
        obj (click.Command or click.Group): Command or group to build help text for

    Yields:
        Text or Markdown: Multiple styled objects (depreciated, usage)
    """
    # Prepend deprecated status
    if obj.deprecated:
        yield Text(config.DEPRECATED_STRING, style=config.STYLE_DEPRECATED)

    # Fetch and dedent the help text
    help_text = inspect.cleandoc(obj.help)  # type: ignore

    # Trim off anything that comes after \f on its own line
    help_text = help_text.partition("\f")[0]

    # Get the first paragraph
    first_line = help_text.split("\n\n")[0]
    # Remove single linebreaks
    if not config.USE_MARKDOWN and not first_line.startswith("\b"):
        first_line = first_line.replace("\n", " ")
    yield _make_rich_rext(first_line.strip(), config=config, style=config.STYLE_HELPTEXT_FIRST_LINE)

    # Get remaining lines, remove single line breaks and format as dim
    remaining_paragraphs = help_text.split("\n\n")[1:]
    if len(remaining_paragraphs) > 0:
        if not config.USE_MARKDOWN:
            # Remove single linebreaks
            remaining_paragraphs = [
                x.replace("\n", " ").strip() if not x.startswith("\b") else "{}\n".format(x.strip("\b\n"))
                for x in remaining_paragraphs
            ]
            # Join back together
            remaining_lines = "\n".join(remaining_paragraphs)
        else:
            # Join with double linebreaks if markdown
            remaining_lines = "\n\n".join(remaining_paragraphs)

        yield _make_rich_rext(remaining_lines, config=config, style=config.STYLE_HELPTEXT)


# flake8: noqa: C901
def _get_parameter_help(
    param: Union[click.Option, click.Argument, click.Parameter], ctx: click.Context, config: RichConfig
) -> rich.columns.Columns:
    """Build primary help text for a click option or argument.

    Returns the prose help text for an option or argument, rendered either
    as a Rich Text object or as Markdown.
    Additional elements are appended to show the default and required status if applicable.

    Args:
        param (click.Option or click.Argument): Option or argument to build help text for
        ctx (click.Context): Click Context object

    Returns:
        Columns: A columns element with multiple styled objects (help, default, required)
    """
    items: List[Union[Markdown, Text, Any]] = []

    # Get the environment variable first
    envvar = getattr(param, "envvar", None)
    # https://github.com/pallets/click/blob/0aec1168ac591e159baf6f61026d6ae322c53aaf/src/click/core.py#L2720-L2726
    if envvar is None:
        if (
            getattr(param, "allow_from_autoenv", None)
            and getattr(ctx, "auto_envvar_prefix", None) is not None
            and param.name is not None
        ):
            envvar = f"{ctx.auto_envvar_prefix}_{param.name.upper()}"
    if envvar is not None:
        envvar = ", ".join(param.envvar) if type(envvar) is list else envvar  # type: ignore

    # Environment variable BEFORE help text
    if getattr(param, "show_envvar", None) and config.OPTION_ENVVAR_FIRST and envvar is not None:
        items.append(Text(config.ENVVAR_STRING.format(envvar), style=config.STYLE_OPTION_ENVVAR))

    # Main help text
    if getattr(param, "help", ""):
        paragraphs = param.help.split("\n\n")  # type: ignore
        # Remove single linebreaks
        if not config.USE_MARKDOWN:
            paragraphs = [
                x.replace("\n", " ").strip() if not x.startswith("\b") else "{}\n".format(x.strip("\b\n"))
                for x in paragraphs
            ]
        items.append(_make_rich_rext("\n".join(paragraphs).strip(), config=config, style=config.STYLE_OPTION_HELP))

    # Append metavar if requested
    if config.APPEND_METAVARS_HELP:
        metavar_str = param.make_metavar()
        # Do it ourselves if this is a positional argument
        if isinstance(param, click.core.Argument) and re.match(
            rf"\[?{getattr(param, 'name', '').upper()}]?", metavar_str
        ):
            metavar_str = param.type.name.upper()
        # Skip booleans
        if metavar_str != "BOOLEAN" or isinstance(param, click.core.Argument):
            metavar_str = metavar_str.replace("[", "").replace("]", "")
            items.append(
                Text(
                    config.APPEND_METAVARS_HELP_STRING.format(metavar_str),
                    style=config.STYLE_METAVAR_APPEND,
                    overflow="fold",
                )
            )

    # Environment variable AFTER help text
    if getattr(param, "show_envvar", None) and not config.OPTION_ENVVAR_FIRST and envvar is not None:
        items.append(Text(config.ENVVAR_STRING.format(envvar), style=config.STYLE_OPTION_ENVVAR))

    # Default value
    if getattr(param, "show_default", None):
        # param.default is the value, but click is a bit clever in choosing what to show here
        # eg. --debug/--no-debug, default=False will show up as [default: no-debug] instead of [default: False]
        # To avoid duplicating loads of code, let's just pull out the string from click with a regex
        # Example outputs from param.get_help_record(ctx)[-1] are:
        #     [default: foo]
        #     [env var: EMAIL, EMAIL_ADDRESS; default: foo]
        default_str_match = re.search(r"\[(?:.+; )?default: (.*)\]", typing.cast(list, param.get_help_record(ctx))[-1])
        if default_str_match:
            # Don't show the required string, as we show that afterwards anyway
            default_str = default_str_match.group(1).replace("; required", "")
            items.append(
                Text(
                    config.DEFAULT_STRING.format(default_str),
                    style=config.STYLE_OPTION_DEFAULT,
                )
            )

    # Required long?
    if config.APPEND_METAVARS_REQUIRED and param.required:
        items.append(Text(config.REQUIRED_LONG_STRING, style=config.STYLE_REQUIRED_LONG))

    # Use Columns - this allows us to group different renderable types
    # (Text, Markdown) onto a single line.
    return Columns(items)


# flake8: noqa: C901
def _make_command_help(config: RichConfig, help_text: str = "") -> Union[rich.text.Text, rich.markdown.Markdown]:
    """Build click help text for a click group command.

    That is, when calling help on groups with multiple subcommands
    (not the main help text when calling the subcommand help).

    Returns the first paragraph of help text for a command, rendered either as a
    Rich Text object or as Markdown.
    Ignores single newlines as paragraph markers, looks for double only.

    Args:
        help_text (str): Help text

    Returns:
        Text or Markdown: Styled object
    """
    paragraphs = inspect.cleandoc(help_text).split("\n\n")
    # Remove single linebreaks
    if not config.USE_MARKDOWN and not paragraphs[0].startswith("\b"):
        paragraphs[0] = paragraphs[0].replace("\n", " ")
    elif paragraphs[0].startswith("\b"):
        paragraphs[0] = paragraphs[0].replace("\b\n", "")
    return _make_rich_rext(paragraphs[0].strip(), config=config, style=config.STYLE_OPTION_HELP)


def rich_format_help(
    obj: Union[click.Command, click.Group],
    ctx: click.Context,
    formatter: click.HelpFormatter,
    config: RichConfig,
    console: Console = None,
) -> None:
    """Print nicely formatted help text using rich.

    Based on original code from rich-click, by @willmcgugan.
    https://github.com/Textualize/rich-cli/blob/8a2767c7a340715fc6fbf4930ace717b9b2fc5e5/src/rich_cli/__main__.py#L162-L236

    Replacement for the click function format_help().
    Takes a command or group and builds the help text output.

    :param formatter: Click Format Helper
    :param obj: Command or group to build help text for
    :param ctx: Click Context object
    :param config: Rich Config
    :param console: Rich Console
    """
    if not console:
        console = get_rich_console(config=config)

    # Header logo if we have it
    if config.HEADER_LOGO:
        console.print(config.HEADER_LOGO)

    # Header text if we have it
    if config.HEADER_TEXT:
        console.print(
            Padding(_make_rich_rext(config.HEADER_TEXT, config=config, style=config.STYLE_HEADER_TEXT), (1, 1, 0, 1))
        )

    # Print usage
    console.print(Padding(highlighter(obj.get_usage(ctx)), 1), style=config.STYLE_USAGE_COMMAND)

    # Print command / group help if we have some
    if obj.help:
        # Print with some padding
        console.print(
            Padding(
                Align(_get_help_text(obj, config=config), pad=False),
                (0, 1, 1, 1),
            )
        )

    # Look through OPTION_GROUPS for this command
    # stick anything unmatched into a default group at the end
    option_groups = config.OPTION_GROUPS.get(ctx.command_path, []).copy()
    option_groups.append({"options": []})
    argument_group_options = []

    for param in obj.get_params(ctx):

        # Skip positional arguments - they don't have opts or helptext and are covered in usage
        # See https://click.palletsprojects.com/en/8.0.x/documentation/#documenting-arguments
        if isinstance(param, click.core.Argument) and not config.SHOW_ARGUMENTS:
            continue

        # Skip if option is hidden
        if getattr(param, "hidden", False):
            continue

        # Already mentioned in a config option group
        for option_group in option_groups:
            if any([opt in option_group.get("options", []) for opt in param.opts]):
                break

        # No break, no mention - add to the default group
        else:
            if isinstance(param, click.core.Argument) and not config.GROUP_ARGUMENTS_OPTIONS:
                argument_group_options.append(param.opts[0])
            else:
                list_of_option_groups: List = option_groups[-1]["options"]  # type: ignore
                list_of_option_groups.append(param.opts[0])

    # If we're not grouping arguments, and we got some, prepend before default options
    if len(argument_group_options) > 0:
        extra_option_group = {"name": config.ARGUMENTS_PANEL_TITLE, "options": argument_group_options}
        option_groups.insert(len(option_groups) - 1, extra_option_group)  # type: ignore

    # Print each option group panel
    for option_group in option_groups:

        options_rows = []
        for opt in option_group.get("options", []):

            # Get the param
            for param in obj.get_params(ctx):
                if any([opt in param.opts]):
                    break
            # Skip if option is not listed in this group
            else:
                continue

            # Short and long form
            opt_long_strs = []
            opt_short_strs = []
            for idx, opt in enumerate(param.opts):
                opt_str = opt
                try:
                    opt_str += "/" + param.secondary_opts[idx]
                except IndexError:
                    pass

                if isinstance(param, click.core.Argument):
                    opt_long_strs.append(opt_str.upper())
                elif "--" in opt:
                    opt_long_strs.append(opt_str)
                else:
                    opt_short_strs.append(opt_str)

            # Column for a metavar, if we have one
            metavar = Text(style=config.STYLE_METAVAR, overflow="fold")
            metavar_str = param.make_metavar()

            # Do it ourselves if this is a positional argument
            if isinstance(param, click.core.Argument) and re.match(
                rf"\[?{getattr(param, 'name', '').upper()}]?", metavar_str
            ):
                metavar_str = param.type.name.upper()

            # Skip booleans and choices (handled above)
            if metavar_str != "BOOLEAN" or isinstance(param, click.core.Argument):
                metavar.append(metavar_str)

            # Range - from
            # https://github.com/pallets/click/blob/c63c70dabd3f86ca68678b4f00951f78f52d0270/src/click/core.py#L2698-L2706  # noqa: E501
            try:
                # skip count with default range type
                if isinstance(param.type, click.types._NumberRangeBase) and not (
                    getattr(param, "count", 0) and param.type.min == 0 and param.type.max is None
                ):
                    range_str = param.type._describe_range()
                    if range_str:
                        metavar.append(config.RANGE_STRING.format(range_str))
            except AttributeError:
                # click.types._NumberRangeBase is only in Click 8x onwards
                pass

            # Required asterisk
            required: Text = Text("")
            if param.required:
                required = Text(config.REQUIRED_SHORT_STRING, style=config.STYLE_REQUIRED_SHORT)

            # Highlighter to make [ | ] and <> dim
            class MetavarHighlighter(RegexHighlighter):
                highlights = [
                    r"^(?P<metavar_sep>(\[|<))",
                    r"(?P<metavar_sep>\|)",
                    r"(?P<metavar_sep>(\]|>)$)",
                ]

            metavar_highlighter = MetavarHighlighter()

            rows = [
                required,
                highlighter(highlighter(",".join(opt_long_strs))),
                highlighter(highlighter(",".join(opt_short_strs))),
                metavar_highlighter(metavar),
                _get_parameter_help(param, ctx, config=config),
            ]

            # Remove metavar if specified in config
            if not config.SHOW_METAVARS_COLUMN:
                rows.pop(3)

            options_rows.append(rows)

        if len(options_rows) > 0:
            t_styles = {
                "show_lines": config.STYLE_OPTIONS_TABLE_SHOW_LINES,
                "leading": config.STYLE_OPTIONS_TABLE_LEADING,
                "box": config.STYLE_OPTIONS_TABLE_BOX,
                "border_style": config.STYLE_OPTIONS_TABLE_BORDER_STYLE,
                "row_styles": config.STYLE_OPTIONS_TABLE_ROW_STYLES,
                "pad_edge": config.STYLE_OPTIONS_TABLE_PAD_EDGE,
                "padding": config.STYLE_OPTIONS_TABLE_PADDING,
            }
            t_styles.update(option_group.get("table_styles", {}))  # type: ignore
            box_style = getattr(box, t_styles.pop("box"), None)  # type: ignore
            options_table = Table(
                highlight=True,
                show_header=False,
                expand=True,
                box=box_style,
                **t_styles,  # type: ignore
            )
            # Strip the required column if none are required
            if all([x[0] == "" for x in options_rows]):
                options_rows = [x[1:] for x in options_rows]
            for row in options_rows:
                options_table.add_row(*row)  # type: ignore
            console.print(
                Panel(
                    options_table,
                    border_style=config.STYLE_OPTIONS_PANEL_BORDER,
                    title=option_group.get("name", config.OPTIONS_PANEL_TITLE),  # type: ignore
                    title_align=config.ALIGN_OPTIONS_PANEL,  # type: ignore
                )
            )

    #
    # Groups only:
    # List click command groups
    #
    if hasattr(obj, "list_commands"):
        # Look through COMMAND_GROUPS for this command
        # stick anything unmatched into a default group at the end
        cmd_groups = config.COMMAND_GROUPS.get(ctx.command_path, []).copy()
        cmd_groups.append({"commands": []})
        for command in obj.list_commands(ctx):  # type: ignore
            for cmd_group in cmd_groups:
                if command in cmd_group.get("commands", []):
                    break
            else:
                commands: List = cmd_groups[-1]["commands"]  # type: ignore
                commands.append(command)

        # Print each command group panel
        for cmd_group in cmd_groups:
            t_styles = {
                "show_lines": config.STYLE_COMMANDS_TABLE_SHOW_LINES,
                "leading": config.STYLE_COMMANDS_TABLE_LEADING,
                "box": config.STYLE_COMMANDS_TABLE_BOX,
                "border_style": config.STYLE_COMMANDS_TABLE_BORDER_STYLE,
                "row_styles": config.STYLE_COMMANDS_TABLE_ROW_STYLES,
                "pad_edge": config.STYLE_COMMANDS_TABLE_PAD_EDGE,
                "padding": config.STYLE_COMMANDS_TABLE_PADDING,
            }
            t_styles.update(cmd_group.get("table_styles", {}))  # type: ignore
            box_style = getattr(box, t_styles.pop("box"), None)  # type: ignore

            commands_table = Table(
                highlight=False,
                show_header=False,
                expand=True,
                box=box_style,
                **t_styles,  # type: ignore
            )
            # Define formatting in first column, as commands don't match highlighter regex
            commands_table.add_column(style="bold cyan", no_wrap=True)
            for command in cmd_group.get("commands", []):
                # Skip if command does not exist
                if command not in obj.list_commands(ctx):  # type: ignore
                    continue
                cmd = obj.get_command(ctx, command)  # type: ignore
                if cmd.hidden:
                    continue
                # Use the truncated short text as with vanilla text if requested
                if config.USE_CLICK_SHORT_HELP:
                    help_text = cmd.get_short_help_str()
                else:
                    # Use short_help function argument if used, or the full help
                    help_text = cmd.short_help or cmd.help or ""
                commands_table.add_row(command, _make_command_help(config=config, help_text=help_text))
            if commands_table.row_count > 0:
                console.print(
                    Panel(
                        commands_table,
                        border_style=config.STYLE_COMMANDS_PANEL_BORDER,
                        title=cmd_group.get("name", config.COMMANDS_PANEL_TITLE),  # type: ignore
                        title_align=config.ALIGN_COMMANDS_PANEL,  # type: ignore
                    )
                )

    # Epilogue if we have it
    if obj.epilog:
        # Remove single linebreaks, replace double with single
        lines = obj.epilog.split("\n\n")
        epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
        console.print(Padding(Align(highlighter(epilogue), pad=False), 1))

    # Footer text if we have it
    if config.FOOTER_TEXT:
        console.print(
            Padding(_make_rich_rext(config.FOOTER_TEXT, config=config, style=config.STYLE_FOOTER_TEXT), (1, 1, 0, 1))
        )


def rich_format_error(self: click.ClickException, config: RichConfig, console: Console = None):
    """Print richly formatted click errors.

    Called by custom exception handler to print richly formatted click errors.
    Mimics original click.ClickException.echo() function but with rich formatting.

    Args:
        click.ClickException: Click exception to format.
    """
    if not console:
        console = get_rich_console(config=config)
    if getattr(self, "ctx", None) is not None:
        console.print(self.ctx.get_usage())  # type: ignore
    if config.ERRORS_SUGGESTION:
        console.print(config.ERRORS_SUGGESTION, style=config.STYLE_ERRORS_SUGGESTION)
    elif (
        config.ERRORS_SUGGESTION is None
        and getattr(self, "ctx", None) is not None
        and self.ctx.command.get_help_option(self.ctx) is not None  # type: ignore
    ):
        console.print(
            "Try [blue]'{command} {option}'[/] for help.".format(
                command=self.ctx.command_path, option=self.ctx.help_option_names[0]  # type: ignore
            ),
            style=config.STYLE_ERRORS_SUGGESTION,
        )

    console.print(
        Panel(
            highlighter(self.format_message()),
            border_style=config.STYLE_ERRORS_PANEL_BORDER,
            title=config.ERRORS_PANEL_TITLE,
            title_align=config.ALIGN_ERRORS_PANEL,  # type: ignore
        )
    )
    if config.ERRORS_EPILOGUE:
        console.print(config.ERRORS_EPILOGUE)


def rich_abort_error(config: RichConfig, console: Console = None) -> None:
    """Print richly formatted abort error."""
    if not console:
        console = get_rich_console(config=config)
    console.print(config.ABORTED_TEXT, style=config.STYLE_ABORTED)



from dataclasses import dataclass as new_dataclass
from dataclasses import fields
from typing import Any, Callable, Optional, Type, Union

from cornflakes.decorator._add_dataclass_slots import add_slots
from cornflakes.decorator._types import DataclassProtocol
from cornflakes.decorator.config.dict import to_dict
from cornflakes.decorator.config.ini import to_ini, to_ini_bytes
from cornflakes.decorator.config.tuple import to_tuple
from cornflakes.decorator.config.yaml import to_yaml, to_yaml_bytes
from cornflakes.decorator.dataclass._enforce_types import enforce_types as enforce
from cornflakes.decorator.dataclass._field import Field


def dataclass(
    cls=None,
    dict_factory: Optional[Callable] = None,
    tuple_factory: Optional[Callable] = None,
    eval_env: bool = False,
    validate: bool = False,
    **kwargs
) -> Union[DataclassProtocol, Callable[..., DataclassProtocol]]:
    """Wrapper around built-in dataclasses dataclass."""

    def wrapper(w_cls: Type[Any]) -> DataclassProtocol:
        dataclass_fields = {
            obj_name: getattr(w_cls, obj_name)
            for obj_name in dir(w_cls)
            if isinstance(getattr(w_cls, obj_name), Field) and hasattr(getattr(w_cls, obj_name), "alias")
        }
        has_add_slots = (
            kwargs.pop("slots", False)
            if "slots" in kwargs and "slots" not in new_dataclass.__code__.co_varnames
            else False
        )
        dc_cls: Union[DataclassProtocol, Any] = new_dataclass(w_cls, **kwargs)
        if has_add_slots:
            dc_cls = add_slots(dc_cls)
        dc_cls.__dataclass_fields__.update(dataclass_fields)
        dc_cls.__dict_factory__ = dict_factory or dict
        dc_cls.__tuple_factory__ = tuple_factory or tuple
        dc_cls.__eval_env__ = eval_env

        dc_cls.__ignored_slots__ = [f.name for f in fields(dc_cls) if getattr(f, "ignore", False)]

        dc_cls.to_dict = to_dict
        dc_cls.to_tuple = to_tuple
        dc_cls.to_ini = to_ini
        dc_cls.to_yaml = to_yaml
        dc_cls.to_yaml_bytes = to_yaml_bytes
        dc_cls.to_ini_bytes = to_ini_bytes

        if validate:
            dc_cls = enforce(dc_cls, validate)
        return dc_cls

    if cls:
        return wrapper(cls)
    return wrapper



"""cornflakes decorator module.
__________________________________
"""  # noqa: RST303 D205
from cornflakes.common import patch_module
from cornflakes.decorator.config import config, config_group
from cornflakes.decorator._add_dataclass_slots import add_slots
from cornflakes.decorator._click_cli import click_cli
from cornflakes.decorator.dataclass import field, dataclass
from cornflakes.decorator._types import ConfigArguments, Loader

config_field = field


__all__ = [
    "config",
    "config_group",
    "config_field",
    "add_slots",
    "click_cli",
    "dataclass",
    "field",
    "Loader",
    "ConfigArguments",
]

patch_module(globals())



from dataclasses import asdict, fields, is_dataclass
import os
from typing import Any, Callable, Dict, List, Optional, Union

from cornflakes.decorator._types import Config
from cornflakes.decorator.config._helper import dict_factory
from cornflakes.decorator.config._load_config import create_file_loader


def to_dict(self) -> Any:
    """Method to convert Dataclass with slots to dict."""
    if not is_dataclass(self):
        return self
    new_dict = asdict(self, dict_factory=dict_factory(self))
    if not (
        isinstance(new_dict, dict)
        or any([is_dataclass(f.type) or f.default_factory == list or isinstance(f.default, list) for f in dc_fields])
        if (dc_fields := fields(self))
        else True
    ):
        return new_dict
    for f in dc_fields:
        if is_dataclass(value := getattr(self, f.name)):
            new_dict.update({f.name: value.to_dict()})
        if isinstance(value, list) or isinstance(value, tuple):
            value = list(value)  # if tuple cast  to list
            for idx, sub_value in enumerate(value):
                if is_dataclass(sub_value):
                    value[idx] = sub_value.to_dict()
            new_dict.update({f.name: value})
    return new_dict


def create_dict_file_loader(
    cls: Config,
) -> Callable[..., Dict[str, Optional[Union[Config, List[Config]]]]]:
    """Method to create file loader for ini files."""

    def from_dict(*args, config_dict=None, **kwargs) -> Dict[str, Optional[Union[Config, List[Config]]]]:
        if not config_dict:
            config_dict = {}
        default_kwargs = {}
        if cls.__eval_env__:
            default_kwargs.update(
                {key: os.environ[key] for key in cls.__dataclass_fields__.keys() if key in os.environ.keys()}
            )
        default_kwargs.update(kwargs)
        return create_file_loader(cls=cls)(*args, config_dict=config_dict, **default_kwargs)

    return from_dict


# def create_dict_group_loader(
#     cls=None,
# ) -> Callable[..., ConfigGroup]:
#     """Method to create file loader for ini files."""
#
#     def from_dict(*args, config_dict=None, **kwargs) -> ConfigGroup:
#         if not config_dict:
#             config_dict = {}
#         default_kwargs = {}
#         if cls.__eval_env__:
#             default_kwargs.update({key: os.environ[key] for key in cls.__dataclass_fields__.keys()
#                                    if key in os.environ.keys()})
#         default_kwargs.update(kwargs)
#         return create_group_loader(cls=cls)(*args, config_dict=config_dict, **default_kwargs)
#
#     return from_dict



"""cornflakes config generation."""
from cornflakes.common import patch_module
from cornflakes.builder._generate_config_module import generate_group_module

__all__ = ["generate_group_module"]

patch_module(globals())



from enum import Enum
from typing import Any, Callable, ClassVar, Dict, List, Optional, Protocol, TypeVar, Union

_T = TypeVar("_T")


class _WithoutDefault:
    pass


WITHOUT_DEFAULT = _WithoutDefault()


class Loader(Enum):
    """Config Loader Enums."""

    INI_LOADER = "from_ini"
    YAML_LOADER = "from_yaml"
    DICT_LOADER = "from_dict"
    FILE_LOADER = "from_file"


class ConfigArguments(Enum):
    """Config Arguments Enums."""

    files = "files"
    sections = "sections"
    use_regex = "use_regex"
    is_list = "is_list"
    default_loader = "default_loader"
    allow_empty = "allow_empty"
    filter_function = "filter_function"


ConfigArgument = Optional[Optional[Union[Dict[Optional[str], Union[List[str], str]], List[str], str]]]


class LoaderMethod(Protocol):
    """Config loader method protocol."""

    @staticmethod
    def __call__(
        files: ConfigArgument = None,
        sections: ConfigArgument = None,
        keys: ConfigArgument = None,
        defaults: ConfigArgument = None,
        eval_env: bool = False,
        *args,
        **kwargs
    ) -> Any:
        """Config loader method protocol."""
        ...


class DataclassProtocol(Protocol):
    """Dataclass protocol custom implementation."""

    # stdlib attributes
    __dataclass_fields__: ClassVar[Dict[str, Any]]
    __dataclass_params__: ClassVar[Dict]  # in reality `dataclasses._DataclassParams`
    __annotations__: ClassVar[Dict]
    __post_init__: ClassVar[Optional[Callable[..., None]]]
    # _FIELDS: ClassVar[]
    __args__: ClassVar[List[Any]]
    # custom
    __eval_env__: bool = None

    def __call__(self, *args, **kwargs) -> Any:
        """Call Function."""
        ...


class Config(Protocol):
    """Config Protocol Type."""

    # stdlib attributes
    __dataclass_fields__: ClassVar[Dict[str, Any]]
    __dataclass_params__: ClassVar[Dict]  # in reality `dataclasses._DataclassParams`
    __annotations__: ClassVar[Dict]
    __post_init__: ClassVar[Optional[Callable[..., None]]]
    # _FIELDS: ClassVar[]
    __args__: ClassVar[List[Any]]
    # custom
    __eval_env__: bool = None

    __config_sections__: ClassVar[ConfigArgument]
    __config_files__: ClassVar[ConfigArgument]
    __ignored_slots__: ClassVar[List[str]]
    __multi_config__: ClassVar[bool]
    __config_list__: ClassVar[bool]
    __allow_empty_config__: ClassVar[bool]

    def __call__(self, *args, **kwargs) -> Any:
        """Call Function."""
        ...

    def __config_filter_function__(self) -> bool:
        """Callback function to filter config."""
        ...

    def to_dict(self, *args, **kwargs) -> Any:
        """Parse config to dict."""
        ...

    def to_tuple(self, *args, **kwargs) -> Any:
        """Parse config to tuple."""
        ...

    def to_ini(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to ini file / bytes."""
        ...

    def to_yaml(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to yaml file / bytes."""
        ...

    def to_yaml_bytes(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to yaml bytes."""
        ...

    def to_ini_bytes(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to ini bytes."""
        ...

    from_ini: LoaderMethod
    from_yaml: LoaderMethod
    from_dict: LoaderMethod
    from_file: LoaderMethod


class ConfigGroup(Protocol):
    """ConfigGroup Protocol Type."""

    # stdlib attributes
    __dataclass_fields__: ClassVar[Dict[str, Any]]
    __dataclass_params__: ClassVar[Dict]  # in reality `dataclasses._DataclassParams`
    __annotations__: ClassVar[Dict]
    __post_init__: ClassVar[Optional[Callable[..., None]]]
    # _FIELDS: ClassVar[]
    __args__: ClassVar[List[Any]]
    # custom
    __eval_env__: bool = None

    __config_files__: ClassVar[ConfigArgument]
    __multi_config__: ClassVar[bool]
    __config_list__: ClassVar[bool]
    __allow_empty_config__: ClassVar[bool]

    def __call__(self, *args, **kwargs) -> Any:
        """Call Function."""
        ...

    def __config_filter_function__(self) -> bool:
        """Callback function to filter config."""
        ...

    def to_dict(self, *args, **kwargs) -> Any:
        """Parse config to dict."""
        ...

    def to_tuple(self, *args, **kwargs) -> Any:
        """Parse config to tuple."""
        ...

    def to_ini(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to ini file / bytes."""
        ...

    def to_yaml(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to yaml file / bytes."""
        ...

    def to_yaml_bytes(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to yaml bytes."""
        ...

    def to_ini_bytes(self, *args, **kwargs) -> Optional[bytearray]:
        """Parse config to ini bytes."""

    ...

    from_ini: LoaderMethod
    from_yaml: LoaderMethod
    from_dict: LoaderMethod
    from_file: LoaderMethod



import re
from typing import Any, Optional, Union

from cornflakes.decorator._types import Config, ConfigArgument, ConfigGroup, DataclassProtocol, Loader


def is_config(cls):
    """Method to return flag that class is a config class."""
    return hasattr(cls, "__config_sections__")


def is_group(cls):
    """Method to return flag that class is a config group class."""
    return not is_config(cls) and hasattr(cls, "__config_files__")


def config_files(cls) -> ConfigArgument:
    """Method to return class __config_files__."""
    return getattr(cls, "__config_files__", [])


def dict_factory(cls: Union[Config, ConfigGroup, DataclassProtocol]) -> Any:
    """Method to return class __dict_factory__."""
    return getattr(cls, "__dict_factory__", dict)


def normalized_class_name(cls):
    """Method to return class name normalized."""
    return re.sub(r"([a-z])([A-Z])", "\\1_\\2", cls.__name__).lower()


def tuple_factory(cls: Union[Config, ConfigGroup, DataclassProtocol]) -> Any:
    """Method to return class __tuple_factory__."""
    return getattr(cls, "__tuple_factory__", tuple)


def is_config_list(cls):
    """Method to return flag that the object is a list of configs."""
    return getattr(cls, "__config_list__", False) or (
        hasattr(cls, "__args__") and getattr(cls.__args__[0], "__config_list__", False)
    )


def get_not_ignored_slots(cls):
    """Method to return slots that are not ignored fields."""
    return [slot for slot in getattr(cls, "__dataclass_fields__", {}).keys() if slot not in cls.__ignored_slots__]


def is_multi_config(cls):
    """Method to return flag that the config class can be empty."""
    return getattr(cls, "__multi_config__", False)


def is_allow_empty(cls):
    """Method to return flag that the config class can be empty."""
    return getattr(cls, "__allow_empty_config__", False)


def pass_section_name(cls):
    """Method to return flag that the config has section_name in slots, so that the section title is passed in."""
    return "section_name" in cls.__dataclass_fields__.keys()


def get_default_loader(files: Optional[list] = None):
    """Method to get the default loader from filenames."""
    return (
        Loader.DICT_LOADER
        if not files
        else Loader.INI_LOADER
        if files[0][-3:] == "ini"
        else Loader.YAML_LOADER
        if files[0][-3:] == "yaml"
        else Loader.DICT_LOADER
    )



"""Automatically generated Default Config.

!!!This files is autogenerated, please do not modify manually.!!!
!!!Change the file default-htw-logger.ini or the default values in the base-classes of htw_logger.config!!!
!!!IF htw_logger.config not exists run ```make install``` before!!!
"""
from cornflakes import config_group


@config_group
class Config:
    """Main config class of the module."""

    # modules


__all__ = ["Config"]



from collections import OrderedDict
from dataclasses import fields
from functools import partial
import logging
import re
from typing import Any, Callable, Dict, List, Optional, Union

from cornflakes import ini_load
from cornflakes.decorator._types import WITHOUT_DEFAULT, Config, LoaderMethod
from cornflakes.decorator.config._helper import (
    is_config_list,
    is_multi_config,
    normalized_class_name,
    pass_section_name,
)


def _none_omit(obj: list):
    return [v for v in obj if v is not None]


def _default_filter_method(x: Any):
    return True


def create_file_loader(  # noqa: C901
    cls: Config,
    loader: LoaderMethod = ini_load,  # type: ignore
):
    """Config decorator to parse Ini Files and implements from_file method to config-classes.

    :param cls: Config class
    :param loader: Config Loader (ini_load, yaml_load)

    :returns: wrapped class or the wrapper itself with the custom default arguments if the config class is not
    """
    required_keys = [f.name for f in fields(cls) if f.default_factory == WITHOUT_DEFAULT]  # type: ignore

    keys = {key: getattr(value, "alias", key) or key for key, value in list(cls.__dataclass_fields__.items())}

    def create_config(config: dict, allow_empty=None, filter_function=_default_filter_method, **cls_kwargs):
        if not config and allow_empty:
            return
        config.update(cls_kwargs)
        error_args = [key for key in config if key not in cls.__dataclass_fields__]
        if error_args:
            logging.warning(f"Some variables in **{cls.__name__}** have no annotation or are not defined!")
            logging.warning(f"Please check Args: {error_args}")

        #  config_instance
        config_instance = cls(**{key: value for key, value in config.items() if key in cls.__dataclass_fields__})
        if filter_function(config_instance):
            return config_instance

    def _check_required_fields(config_dict):
        return {
            section: config
            for section, config in config_dict.items()
            if not any([True for key in required_keys if key not in config.keys()])
        }

    def _rename_default_section(config_dict):
        if None not in config_dict:
            return config_dict
        config_dict["default"] = config_dict.pop(None)
        return config_dict

    def from_file(
        files: Optional[Union[List[str], str]] = None,
        sections: Optional[Union[List[str], str]] = None,
        config_dict: Optional[Dict[str, Any]] = None,
        filter_function: Optional[Callable[[Config], bool]] = None,
        eval_env: Optional[bool] = None,
        allow_empty: Optional[bool] = None,
        **slot_kwargs,
    ):
        """Config parser from ini files.

        :param files: Default config files
        :param sections: Default config sections
        :param config_dict: Config dictionary to pass already loaded configs
        :param slot_kwargs: Default configs to overwrite passed class
        :param filter_function: Optional filter method for config
        :param eval_env: Flag to evaluate environment variables into default values.
        :param allow_empty: Flag that allows empty config result -> e.g. emtpy list

        :returns: Nested Lists of Config Classes
        """
        if not sections:
            sections = cls.__config_sections__
        if not files:
            files = cls.__config_files__
        if not filter_function:
            filter_function = cls.__config_filter_function__ or _default_filter_method
        if not eval_env:
            eval_env = cls.__eval_env__
        if not allow_empty:
            allow_empty = cls.__allow_empty_config__

        _create_config = partial(create_config, allow_empty=allow_empty, filter_function=filter_function)

        pass_sections = pass_section_name(cls)

        def get_section_kwargs(section):
            return {**slot_kwargs, **({"section_name": section} if pass_sections else {})}

        if not is_multi_config(cls) and isinstance(sections, str):
            logging.debug(f"Load ini from file: {files} - section: {sections} for config {cls.__name__}")

            if not config_dict:
                config_dict = OrderedDict(
                    loader(files={None: files}, sections=sections, keys=keys, defaults=None, eval_env=eval_env)
                )
                config_dict = _check_required_fields(config_dict)
                config_dict = _rename_default_section(config_dict)
                logging.debug(f"Read config with sections: {config_dict.keys()}")

            if not sections and config_dict.keys():
                sections = config_dict.popitem()[0] or normalized_class_name(cls)

            config = _create_config(config_dict.get(sections, {}), **get_section_kwargs(sections))
            if not config:
                return {sections: _create_config({}, **get_section_kwargs(sections))}
            return {sections: config}

        if not config_dict:
            config_dict = OrderedDict(loader(files={None: files}, sections=None, keys=keys, eval_env=eval_env))
            config_dict = _check_required_fields(config_dict)
            config_dict = _rename_default_section(config_dict)

            logging.debug(f"Read config with sections: {config_dict.keys()}")

        regex = f'({"|".join(sections) if isinstance(sections, list) else sections or ""})'
        logging.debug(f"Load all configs that mach **{regex}**")

        config_dict = {
            section: config for section, config in config_dict.items() if bool(re.match(regex, section or ""))
        }
        if not is_config_list(cls):
            return {
                section: config
                for section, config in {
                    section: _create_config(config_dict.get(section, {}), **slot_kwargs) for section in config_dict
                }.items()
                if config
            } or {
                normalized_class_name(cls): _create_config({}, **slot_kwargs)  # no matches
            }

        return {
            normalized_class_name(cls): (
                list(
                    _none_omit(
                        [
                            _create_config(config_dict.get(section, {}), **get_section_kwargs(section))
                            for section in config_dict
                        ]
                    ),
                )
                or _none_omit([_create_config({}, **slot_kwargs)] * is_config_list(cls))
            )
        }

    return from_file



def patch_module(m):
    """Method to overwrite module variables in a generic way.

    1. Overwrite names from submodules declared in __all__ to parent module.
    2. Overwrite doc_string and adds auto summary with objects defined in __all__.
    """
    for obj in [m[x] for x in m["__all__"]]:
        obj.__module__ = m["__name__"]

    m[
        "__doc__"
    ] = f"""{m["__doc__"]}
.. currentmodule:: {m["__name__"]}

.. autosummary::
   :toctree: _generate

    {'''
    '''.join(m["__all__"])}
"""



"""cornflakes (Top-level package).
____________________________________
"""  # noqa: RST303 D205
from _cornflakes import apply_match, eval_csv, eval_datetime, eval_type, extract_between, ini_load

from cornflakes.common import patch_module
from cornflakes.decorator import config, config_group, add_slots, click_cli, Loader, dataclass, field, config_field
from cornflakes.decorator.dataclass.validator import AnyUrl
from cornflakes.builder import generate_group_module
from cornflakes.parser import yaml_load
from cornflakes.logging import attach_log, setup_logging

__author__ = "Semjon Geist"
__email__ = "semjon.geist@ionos.com"
__version__ = "3.0.5"

__all__ = [
    "click_cli",
    "ini_load",
    "eval_type",
    "eval_datetime",
    "eval_csv",
    "extract_between",
    "apply_match",
    "config",
    "config_group",
    "AnyUrl",
    "add_slots",
    "generate_group_module",
    "yaml_load",
    "attach_log",
    "setup_logging",
    "Loader",
    "patch_module",
    "dataclass",
    "field",
    "config_field",
]

patch_module(globals())



from functools import wraps


def wraps_partial(*partial_args, **partial_kwargs):
    """Wraps a partial a function."""

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            args = [*partial_args, *args[len(partial_args) :]]
            kwargs.update(partial_kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorate



"""cornflakes default validators for config_fields."""
from cornflakes.common import patch_module
from cornflakes.decorator.dataclass.validator.url import AnyUrl

__all__ = ["AnyUrl"]

patch_module(globals())



from decimal import Decimal


def type_to_str(f):
    """Function to convert python object to string -> fix scientific notation for float / Decimal."""
    if isinstance(f, Decimal) or isinstance(f, float):
        return (
            f"0.{'0' * (int(string[(e + 2):]) - 1)}{string[0]}{string[2:e]}"
            if (e := (string := str(f).lower()).find("e")) != -1
            else str(f)
        )
    return str(f)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cornflakes',
 'cornflakes.builder',
 'cornflakes.cli',
 'cornflakes.click',
 'cornflakes.click.options',
 'cornflakes.click.rich',
 'cornflakes.common',
 'cornflakes.decorator',
 'cornflakes.decorator.config',
 'cornflakes.decorator.dataclass',
 'cornflakes.decorator.dataclass.validator',
 'cornflakes.logging',
 'cornflakes.parser']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'click>=8.1.3,<9.0.0',
 'rich-rst>=1.1.7,<2.0.0',
 'rich>=12.6.0,<13.0.0',
 'types-pyyaml>=6.0.12.1,<7.0.0.0']

entry_points = \
{'console_scripts': ['cornflakes = cornflakes.__main__:main']}

setup_kwargs = {
    'name': 'cornflakes',
    'version': '3.0.5',
    'description': 'Create generic any easy to manage Configs for your Project',
    'long_description': '.. image:: assets/cornflakes.png\n   :height: 400 px\n   :width: 400 px\n   :alt: cornflakes logo\n   :align: center\n\n==========\n\n|PyPI| |Python Version| |License| |Read the Docs| |Build| |Tests| |Codecov|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/cornflakes.svg\n   :target: https://pypi.org/project/cornflakes/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/cornflakes\n   :target: https://pypi.org/project/cornflakes\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/github/license/semmjon/cornflakes\n   :target: https://opensource.org/licenses/Apache2.0\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/cornflakes/latest.svg?label=Read%20the%20Docs\n   :target: https://cornflakes.readthedocs.io\n   :alt: Read the documentation at https://cornflakes.readthedocs.io\n.. |Build| image:: https://github.com/semmjon/cornflakes/workflows/Build%20cornflakes%20Package/badge.svg\n   :target: https://github.com/semmjon/cornflakes/actions?workflow=Package\n   :alt: Build Package Status\n.. |Tests| image:: https://github.com/semmjon/cornflakes/workflows/Run%20cornflakes%20Tests/badge.svg\n   :target: https://github.com/semmjon/cornflakes/actions?workflow=Tests\n   :alt: Run Tests Status\n.. |Codecov| image:: https://codecov.io/gh/semmjon/cornflakes/branch/release-1.4.5/graph/badge.svg?token=FY72EIXI82\n   :target: https://codecov.io/gh/semmjon/cornflakes\n   :alt: Codecov\n\n.. code::\n\n   pip install cornflakes\n\n.. code::\n\n    pip install git+https://github.com/semmjon/cornflakes\n\nInformation\n-----------\n\nThis package was created by starting C ++ methods to incorporate into my python implementations.\nTo make things easier for me, lightweight public libraries were included\n(especially to carry out string operations):\n\n* hash-library\n* strtk\n* rapidjson\n\nFeatures:\n~~~~~~~~~\n\nThe following features have currently been implemented:\n    * config management system\n        - based on dataclass\n        - alternative Implementation for pydantic (BaseSettings)\n        - ini support files by a lightweight and fast parser (-> ini_load)\n        - yaml support (based on PyYAML)\n        - environment variables\n        - (future) support json (based on orjson)\n    * command line interface management\n        - method: click_cli (decorator)\n        - based on click and rich\n        - easy to use and start with\n    * eval_type\n        - method to parse strings in python-types e.g. int | bool | timestamp\n    * simple_hmac\n        - vectorized c++ hmac implementation\n    * default_ca_path\n        - python function to find a default ssl / ca certificate path\n\nCurrently, the package is tested for Linux, Mac and Windows\n\nDevelopment\n-----------\n\nPrerequisites\n~~~~~~~~~~~~~\n\n-  A compiler with C++17 support\n-  Pip 10+ or CMake >= 3.4 (or 3.8+ on Windows, which was the first version to support VS 2015)\n-  Python 3.8+\n-  doxygen\n-  cppcheck\n-  clang-tools-extra or clang-tidy\n-  ..\n\nCommands\n~~~~~~~~~~~~\n\nJust clone this repository and pip install. Note the ``--recursive``\noption which is needed for the pybind11 submodule:\n\n.. code::\n\n   git clone --recursive https://gitlab.blubblub.tech/sgeist/cornflakes.git\n\nInstall the package using makefiles:\n\n.. code::\n\n   make install\n\nBuild dist using makefiles:\n\n.. code::\n\n   make dist\n\nRun tests (pytest) using makefiles:\n\n.. code::\n\n   make test\n\n\nRun all tests using makefiles:\n\n.. code::\n\n   make test-all\n\nRun lint using makefiles:\n\n.. code::\n\n   make lint\n\nCreate dev venv:\n\n.. code::\n\n   python -m venv .venv\n   source .venv/bin/activate\n   pip install cookietemple ninja pre-commit poetry\n\nBump Version using cookietemple:\n\n.. code::\n\n   cookietemple bump-version "<version(e.g 0.0.1)>"\n\nRun lint using cookietemple:\n\n.. code::\n\n   cookietemple lint .\n\nInstall pre-commit:\n\n.. code::\n\n   pre-commit install\n\nUpdate pre-commit:\n\n.. code::\n\n   pre-commit update -a\n\nRun pre-commit:\n\n.. code::\n\n   pre-commit run -a\n\nPublish\n~~~~~~~\n\nIts not recommended publish manually (use git-ci or github workflows instead).\n\n.. code::\n\n   make publish\n',
    'author': 'Semjon Geist',
    'author_email': 'semjon.geist@ionos.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sgeist/cornflakes',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4.0.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from inspect import getfile
import logging
from os.path import abspath
import subprocess  # noqa: S404
import sys
from typing import Callable

from cornflakes.click.options._global import global_option


@global_option(
    ["-b", "--background-process"],
    is_flag=True,
    help="Run in Background without console logger.",
)
def bg_process_option(self: Callable[..., None], background_process: bool, *func_args, **func_kwargs):
    """Default Option for running in background."""
    if background_process:
        stdout_file = f"{self.__name__}.log"
        stderr_file = f"{self.__name__}_error.log"
        logging.debug(
            f"Method {self.__name__} is running in background. "
            f"See logs at stdout: {stdout_file}, stderr: {stderr_file}."
        )
        stdout = open(stdout_file, "w")
        stderr = open(stderr_file, "w")

        command = (
            f"import sys; exec(open('"
            f"{abspath(getfile(self))}').read());"
            f"{self.__name__}(*{func_args},**{func_kwargs});"
        )

        logging.debug(f"Python Command: {command}")

        subprocess.Popen(  # noqa: S603
            [
                sys.executable,
                "-c",
                command,
            ],
            stdout=stdout,
            stderr=stderr,
        )
        sys.exit(0)



import os
from typing import Tuple


def default_ca_path() -> str:
    """Find default ca-certificate of the current system.

    :return: system default ca-certificate
    """
    # Possible certificate files; stop after finding one.
    cert_paths: Tuple[str, ...] = (
        "/etc/ssl/certs/ca-certificates.crt",  # Debian/Ubuntu/Gentoo etc.
        "/etc/pki/tls/certs/ca-bundle.crt",  # Fedora/RHEL 6
        "/etc/ssl/ca-bundle.pem",  # OpenSUSE
        "/etc/pki/tls/cacert.pem",  # OpenELEC
        "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",  # CentOS/RHEL 7
        "/etc/ssl/cert.pem",  # Alpine Linux
    )

    # Possible directories with certificate files; stop after successfully
    # reading at least one file from a directory.
    cert_directories: Tuple[str, ...] = (
        "/etc/ssl/certs",  # SLES10/SLES11, https://golang.org/issue/12139
        "/etc/pki/tls/certs",  # Fedora/RHEL
        "/system/etc/security/cacerts",  # Android
    )

    return next(
        (cert_path for cert_path in cert_paths if os.path.isfile(cert_path) and os.access(cert_path, os.R_OK)),
        next(
            (
                next(
                    (
                        os.path.join(cert_dir, cert_path)
                        for cert_path in os.listdir(cert_dir)
                        if os.access(os.path.join(cert_dir, cert_path), os.R_OK)
                    ),
                    "",
                )
                for cert_dir in cert_directories
                if os.path.exists(cert_dir)
            ),
            "",
        ),
    )



"""cornflakes Logger.
__________________________________
"""  # noqa: RST303 D205
from cornflakes.common import patch_module
from cornflakes.logging._logger import attach_log, setup_logging

__all__ = ["attach_log", "setup_logging"]

patch_module(globals())



"""Click Extension for better CLI.
__________________________________
See referenced Code at https://github.com/ewels/rich-click.git
"""  # noqa: RST303 D205

# from click import *  # noqa: F401, F403
import click

from cornflakes.common import patch_module
from cornflakes.click.rich import RichArg, argument, RichCommand, command, RichGroup, group, RichConfig
from cornflakes.click.options import bg_process_option, verbose_option, global_option, auto_option

# Pach click
click.argument = argument
click.group = group
click.command = command
click.Group = RichGroup
click.Command = RichCommand
click.Argument = RichArg

__all__ = [
    "global_option",
    "verbose_option",
    "bg_process_option",
    "auto_option",
    "group",
    "command",
    "argument",
    "RichGroup",
    "RichCommand",
    "RichConfig",
]

patch_module(globals())



import sys
from typing import Any, List

from click import ClickException, Command, Context, HelpFormatter, Parameter, exceptions

from cornflakes.click.rich._rich_click import rich_abort_error, rich_format_error, rich_format_help
from cornflakes.click.rich._rich_config import RichConfig as RichConfig


class RichCommand(Command):
    """Richly formatted click Command.

    Inherits click.Command and overrides help and error methods
    to print richly formatted output.
    """

    standalone_mode = False
    params: List[Parameter]
    allow_extra_args = True
    allow_interspersed_args = False
    ignore_unknown_options = False
    name = ""
    context_settings: dict

    def callback(self):
        """Callback method with is wrapped over the command."""

    def __init__(self, config: RichConfig = None, *args, **kwargs):
        """Init function of RichGroup with extra config argument."""
        if not config:
            config = RichConfig()
        super().__init__(*args, **kwargs)
        self.config = config
        self.console = None

    def main(self, *args, standalone_mode: bool = True, **kwargs) -> Any:  # noqa: C901
        """Main function of RichGroup."""
        try:
            rv = super().main(*args, standalone_mode=False, **kwargs)  # type: ignore
            if not standalone_mode:
                return rv
        except ClickException as e:
            if not standalone_mode:
                raise
            rich_format_error(e, config=self.config)
            sys.exit(e.exit_code)
        except exceptions.Abort:
            if not standalone_mode:
                raise
            rich_abort_error(config=self.config)
            sys.exit(1)

    def format_help(self, ctx: Context, formatter: HelpFormatter) -> None:
        """Format function of RichGroup."""
        rich_format_help(self, ctx, formatter, config=self.config)



"""Command-line interface."""
from cornflakes.cli import test
from cornflakes.click import RichGroup
from cornflakes.click.options import bg_process_option, verbose_option
from cornflakes.decorator import click_cli


@click_cli(
    OPTION_GROUPS={
        **{
            command: [
                {
                    "name": "Basic Options",
                    "options": [
                        "--name",
                        "--description",
                        "--help",
                        "--log-level",
                        "--log-config",
                        "--version",
                        "--verbose",
                        "--silent",
                        "--background-process",
                    ],
                },
            ]
            for command in ["cornflakes", "cornflakes create"]
        },
    },
    COMMAND_GROUPS={
        "cornflakes": [
            {
                "name": "Commands",
                "commands": [],
            },
        ]
    },
    CONTEXT_SETTINGS={"help_option_names": ["-h", "--help"]},
    HEADER_LOGO="""[blue]
 ██████╗ ██████╗ ██████╗ ███╗   ██╗███████╗██╗      █████╗ ██╗  ██╗███████╗ ███████╗
██╔════╝██╔═══██╗██╔══██╗████╗  ██║██╔════╝██║     ██╔══██╗██║ ██╔╝██╔════╝ ██╔════╝
██║     ██║   ██║██████╔╝██╔██╗ ██║█████╗  ██║     ███████║█████╔╝ █████╗   ███████╗
██║     ██║   ██║██╔══██╗██║╚██╗██║██╔══╝  ██║     ██╔══██║██╔═██╗ ██╔══╝   ╚════██║
╚██████╗╚██████╔╝██║  ██║██║ ╚████║██║     ███████╗██║  ██║██║  ██╗███████╗ ███████║
 ╚═════╝ ╚═════╝ ╚═╝  ╚═╝╚═╝  ╚═══╝╚═╝     ╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝╚══════╝ ╚══════╝""",
    # Banner generated with figlet and figlet-fonts/ANSI\ Shadow.flf
    HEADER_TEXT="Create generic any easy to manage Configs for your Project.",
    GLOBAL_OPTIONS=[verbose_option, bg_process_option],
    VERSION_INFO=True,
)
def main():
    """Main CLI Entrypoint Method."""


if isinstance(main, RichGroup):
    for command in [test]:
        main.add_command(command)
        main.config.COMMAND_GROUPS.get("cornflakes")[0].get("commands").append(command.name)  # type: ignore

__all__ = ["main"]

if __name__ == "__main__":
    main()  # pragma: no cover



"""custom dataclass wrapper."""
# from dataclasses import *  # noqa: F401, F403
from cornflakes.common import patch_module
from cornflakes.decorator.dataclass._dataclass import dataclass
from cornflakes.decorator.dataclass._field import field, Field

__all__ = ["field", "dataclass", "Field"]

patch_module(globals())



import dataclasses
from dataclasses import dataclass, field
from functools import wraps
import inspect
from inspect import Parameter, Signature, signature
from typing import Any, Callable, Dict, List, Optional

_HAS_DEFAULT_FACTORY = getattr(dataclasses, "_HAS_DEFAULT_FACTORY", None)
Empty = getattr(inspect, "_empty", None)


# from cornflakes.decorator.dataclass._dataclass import dataclass as data


def _not_excluded(default):
    return default != _HAS_DEFAULT_FACTORY


def _not_empty(x):
    return x != Empty


def _check_default(default):
    return _not_empty(default) and _not_excluded(default)


def _check_annotation(annotation):
    return _not_empty(annotation)


@dataclass
class KwargsWrapper:
    """KwargsWrapper Class."""

    wrapped_sig: Optional[Signature] = field(default=None, init=False)
    key_names: List[str] = field(default_factory=list, init=False)
    arg_names: List[str] = field(default_factory=list, init=False)
    kwarg_names: List[str] = field(default_factory=list, init=False)
    wrapped: Optional[Callable[..., Any]] = field(default=None)
    key_params: List[Parameter] = field(default_factory=list)
    key_params_no_default: List[Parameter] = field(default_factory=list)
    arg_params: List[Parameter] = field(default_factory=list)
    kwarg_params: List[Parameter] = field(default_factory=list)

    @property
    def _names(self):
        return [*self.key_names, *self.arg_names, *self.kwarg_names]

    @property
    def _passed_names(self):
        return ", ".join(
            [
                *[f"{key}={key}" for key in self.key_names],
                *[f"*{arg}" for arg in self.arg_names],
                *[f"**{arg}" for arg in self.kwarg_names],
            ]
        )

    @property
    def _params(self):
        return [*self.key_params_no_default, *self.arg_params, *self.key_params, *self.kwarg_params]

    @property
    def _params_declaration(self):
        return ", ".join(
            [
                (
                    f'{str(param).split(":", 1)[0].split("=", 1)[0]}'
                    f'{f": wrapped_type_{idx} " * _check_annotation(param.annotation)}'
                    f'{f"=wrapped_default_value_{idx}" * _check_default(param.default)}'
                )
                for idx, param in enumerate(self._params)
            ]
        )

    @property
    def _defaults_dict(self):
        return {
            f"wrapped_default_value_{idx}": param.default
            for idx, param in enumerate(self._params)
            if _check_default(param.default)
        }

    @property
    def _annotations_dict(self):
        return {
            f"wrapped_type_{idx}": param.annotation
            for idx, param in enumerate(self._params)
            if _check_annotation(param.annotation)
        }

    def _update_params(self, parameters):
        for name, param in parameters.items():
            if name not in self._names:
                if _check_default(param.default):
                    self.key_names.append(name)
                    self.key_params.append(param)
                    continue
                if param.kind == Parameter.VAR_POSITIONAL:
                    self.arg_names.append(name)
                    self.arg_params.append(param)
                    continue
                if param.kind == Parameter.VAR_KEYWORD:
                    self.kwarg_names.append(name)
                    self.kwarg_params.append(param)
                    continue
                if _not_excluded(param.default):
                    self.key_names.append(name)
                    self.key_params_no_default.append(param)
                continue

            is_variadic = Parameter.VAR_POSITIONAL or Parameter.VAR_KEYWORD
            param_idx = self._names.index(name)
            # print(f"overwrite {self.params[param_idx].name} at {param_idx} with {name} of {self.names}")
            if is_variadic:
                self._params[param_idx] = Parameter(
                    param.name,
                    kind=param.kind,
                    annotation=param.annotation
                    if _check_annotation(param.annotation)
                    else self._params[param_idx].annotation,
                )
                continue
            self._params[param_idx] = Parameter(
                param.name,
                kind=param.kind,
                default=param.default if _check_default(param.default) else self._params[param_idx].default,
                annotation=param.annotation
                if _check_annotation(param.annotation)
                else self._params[param_idx].annotation,
            )

    def __post_init__(self):
        """Initialize the Class."""
        self.wrapped_sig: Signature = signature(self.wrapped)
        self._update_params(self.wrapped_sig.parameters)

    def wrap(self, wrapper: Callable[..., Any]) -> Callable[..., Any]:
        """Method to return the declaration string."""
        wrapper_sig: Signature = signature(wrapper)
        self._update_params(wrapper_sig.parameters)
        ldict: Dict[str, Any] = {**self._defaults_dict, **self._annotations_dict}
        # print(ldict)
        # print(self.params_declaration)
        # print(self.passed_names)
        wrapper_str = f"""
def wrap_kwargs({self._params_declaration}):
    return wrapper({self._passed_names})
"""
        # print(wrapper_str)
        exec(  # noqa: S102
            wrapper_str,
            locals(),
            ldict,
        )
        return wraps(self.wrapped)(ldict["wrap_kwargs"])


def wrap_kwargs(wrapped) -> Callable[..., Any]:
    """Function Decorator that can update all passed arguments (that can be a variable) of function."""
    kwargs_wrapper = KwargsWrapper(wrapped)

    def wrapper(func):
        return kwargs_wrapper.wrap(func)

    return wrapper
